# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from mct_quantizers.common.constants import FOUND_ONNXRUNTIME, FOUND_ONNXRUNTIME_EXTENSIONS

if FOUND_ONNXRUNTIME and FOUND_ONNXRUNTIME_EXTENSIONS:
    import onnxruntime as ort
    from onnxruntime_extensions import get_library_path

    def get_ort_session_options() -> ort.SessionOptions:
        """
        Returns: Session options for loading onnxruntime inference session
         with custom implementation of onnx ops.
        """
        opt = ort.SessionOptions()
        opt.register_custom_ops_library(get_library_path())
        return opt

else:
    def get_ort_session_options():
        raise Exception('Installing onnxruntime onnxruntime-extensions and is mandatory '
                        'when using get_ort_session_options. '
                        'Could not find a package.')




# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from mct_quantizers.pytorch.quantizers.base_quantizer_autograd_function import BaseQuantizerAutogradFunction


class BaseWeightQuantizerAutogradFunction(BaseQuantizerAutogradFunction):
    """
    Custom autograd function for weights quantizer.
    It provides a way to define a custom forward and symbolic operation
    and currently does not implement a backward operation.
    """

    @staticmethod
    def is_signed():
        return True



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

__version__ = "1.5.2"


from mct_quantizers.common.base_inferable_quantizer import QuantizationTarget, BaseInferableQuantizer, mark_quantizer
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.keras.activation_quantization_holder import KerasActivationQuantizationHolder
from mct_quantizers.pytorch.activation_quantization_holder import PytorchActivationQuantizationHolder
from mct_quantizers.keras.load_model import keras_load_quantized_model
from mct_quantizers.keras.quantize_wrapper import KerasQuantizationWrapper
from mct_quantizers.pytorch.load_model import pytorch_load_quantized_model
from mct_quantizers.pytorch.quantize_wrapper import PytorchQuantizationWrapper

from mct_quantizers.common import constants
from mct_quantizers.keras import quantizers as keras_quantizers
from mct_quantizers.pytorch import quantizers as pytorch_quantizers

from mct_quantizers.pytorch.onnxruntime_session_options import get_ort_session_options





# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, FOUND_ONNXRUNTIME_EXTENSIONS, ONNX_CUSTOM_OP_DOMAIN
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.pytorch.onnxruntime_validations import validate_weight_params

if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizers.base_symmetric_inferable_quantizer import BaseSymmetricInferableQuantizer
    from mct_quantizers.pytorch.quantizer_utils import to_torch_tensor, get_working_device
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.base_weight_quantizer_autograd_function import \
        BaseWeightQuantizerAutogradFunction


    def quantize_sym_weights_torch(input_tensor: torch.Tensor,
                                   num_bits: int,
                                   threshold: float,
                                   per_channel: bool,
                                   channel_axis: int):
        """
           Quantizes the input tensor symmetrically using torch.

           Args:
               input_tensor (torch.Tensor): The input tensor to be quantized.
               num_bits (int): Number of bits to represent the quantized value.
               threshold (float): The quantization threshold.
               per_channel (bool): Quantize input tensor per-channel or per-tensor.
               channel_axis (int): Axis to quantize the tensor in case of per-channel quantization.

           Returns:
               Symmetrically quantized tensor.
        """

        if isinstance(threshold, np.ndarray):
            threshold = torch.tensor(threshold, dtype=torch.float32).to(get_working_device())

        input_tensor = input_tensor
        scale = threshold / (2 ** (num_bits - 1))
        _min, _max = -threshold, threshold - scale

        if per_channel:
            ones = [1] * input_tensor.ndim
            ones[channel_axis] = -1
            new_shape = tuple(ones)
            _min = torch.reshape(_min, new_shape)
            _max = torch.reshape(_max, new_shape)
            scale = torch.reshape(scale, new_shape)

        # Use torch.where to clip the values in x
        clipped_x = torch.where(input_tensor < _min, _min, input_tensor)
        quantized = torch.round(torch.where(input_tensor > _max, _max, clipped_x) / scale) * scale

        return quantized


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.SYMMETRIC],
                    identifier=QuantizerID.INFERABLE)
    class WeightsSymmetricInferableQuantizer(BaseSymmetricInferableQuantizer):
        """
        Class for quantizing weights using a symmetric quantizer.
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.
            """

            super(WeightsSymmetricInferableQuantizer, self).__init__(threshold=threshold,
                                                                     num_bits=num_bits,
                                                                     signed=True)

            if per_channel:
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert len(
                    threshold) >= 1, f'In per-channel quantization threshold should be of length >= 1 but is ' \
                                     f'{len(threshold)}'
            else:
                assert len(
                    threshold) == 1, f'In per-tensor quantization threshold should be of length 1 but is ' \
                                     f'{len(threshold)}'

            self.per_channel = per_channel
            self.channel_axis = channel_axis

            self.scales = to_torch_tensor(self.scales).to(get_working_device())
            self.zero_points = torch.zeros(len(threshold), dtype=torch.int32).to(get_working_device())

        def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """

            if self._use_custom_impl and torch.jit.is_tracing():
                return WeightsSymmetricF.apply(inputs,
                                               self.num_bits,
                                               self.threshold_np,
                                               self.per_channel,
                                               self.channel_axis)

            inputs.requires_grad = False
            if self.per_channel:
                return torch.fake_quantize_per_channel_affine(inputs,
                                                              self.scales,
                                                              self.zero_points,
                                                              axis=self.channel_axis,
                                                              quant_min=self.min_quantized_domain,
                                                              quant_max=self.max_quantized_domain)
            return torch.fake_quantize_per_tensor_affine(inputs,
                                                         self.scales,
                                                         self.zero_points,
                                                         quant_min=self.min_quantized_domain,
                                                         quant_max=self.max_quantized_domain)


    class WeightsSymmetricF(BaseWeightQuantizerAutogradFunction):
        """
        Custom autograd function for symmetric weights quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, num_bits, threshold, per_channel, channel_axis):
            """
             Forward computation function. This method performs the forward computation using
             the given quantize_sym_weights_torch function.

             Args:
                 ctx: An object that can be used to stash information for backward function.
                 input_tensor: The input tensor to be quantized.
                 num_bits: The number of bits to represent the quantized tensor.
                 threshold: The quantization threshold.
                 per_channel: whether to use per-channel quantization
                 channel_axis: Axis of input to apply per-channel quantization on.

             Returns:
                 The quantized tensor.
             """
            return quantize_sym_weights_torch(input_tensor, num_bits, threshold, per_channel, channel_axis)

        @staticmethod
        def symbolic(g, input_tensor, num_bits, threshold, per_channel, channel_axis):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                num_bits: The number of bits to represent the quantized value.
                threshold: The quantization threshold.
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            # When None is passed as channel_axis, the op has no attribute of channel_axis,
            # which creates conflict with the onnxruntime function. For this reason, if we quantize
            # per-tensor and channel_axis is None, we set it to 0.
            if not per_channel and channel_axis is None:
                channel_axis = 0

            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsSymmetricQuantizer", input_tensor,
                        g.op('Constant', value_t=torch.tensor(threshold, dtype=torch.float32)),
                        num_bits_i=num_bits,
                        per_channel_i=int(per_channel),
                        channel_axis_i=channel_axis,
                        signed_i=int(WeightsSymmetricF.is_signed()),
                        **WeightsSymmetricF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())



else:
    class WeightsSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using WeightsSymmetricInferableQuantizer. '
                            'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef
    def quantize_sym_weights_numpy(input_tensor: np.ndarray,
                                   num_bits: int,
                                   threshold: np.ndarray,
                                   per_channel: int,
                                   channel_axis: int=None):
        """
           Quantizes the input tensor symmetrically using numpy.

           Args:
               input_tensor (np.ndarray): The input tensor to be quantized.
               num_bits (int): Number of bits to represent the quantized value.
               threshold (float): The quantization threshold.
               per_channel (bool): Quantize input tensor per-channel or per-tensor.
               channel_axis (int): Axis to quantize the tensor in case of per-channel quantization.

           Returns:
               Symmetrically quantized tensor.
        """
        validate_weight_params(input_tensor=input_tensor,
                               per_channel=per_channel,
                               min_range=-threshold,
                               max_range=threshold,
                               channel_axis=channel_axis)

        scale = threshold / (2 ** (num_bits - 1))
        _min, _max = -threshold, threshold - scale
        if per_channel:
            ones = np.ones(input_tensor.ndim)
            ones[channel_axis] = -1
            new_shape = tuple([int(x) for x in ones])
            _min = np.reshape(_min, new_shape)
            _max = np.reshape(_max, new_shape)
            scale = np.reshape(scale, new_shape)

        # Use torch.where to clip the values in x
        clipped_x = np.where(input_tensor < _min, _min, input_tensor)
        quantized = np.round(np.where(input_tensor > _max, _max, clipped_x) / scale) * scale
        return quantized


    # Add onnx op function to use during onnxruntime WeightsSymmetricQuantizer op inference
    # Using this decorator the op WeightsSymmetricQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsSymmetricQuantizer",
             inputs=[PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={
                 "num_bits": PyCustomOpDef.dt_int64,
                 "per_channel": PyCustomOpDef.dt_int64,
                 "channel_axis": PyCustomOpDef.dt_int64,
             })
    def weight_sym_ort(input_tensor: np.ndarray,
                       threshold: np.ndarray,
                       **kwargs):

        return quantize_sym_weights_numpy(input_tensor,
                                          kwargs["num_bits"],
                                          threshold,
                                          kwargs["per_channel"],
                                          kwargs["channel_axis"]
                                          )



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, LUT_VALUES_BITWIDTH, EPS, ONNX_CUSTOM_OP_DOMAIN, \
    FOUND_ONNXRUNTIME_EXTENSIONS
from mct_quantizers.common.quant_info import QuantizationMethod

if FOUND_TORCH:
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.weights_lut_symmetric_inferable_quantizer import \
    WeightsLUTSymmetricInferableQuantizer, WeightsLUTSymmetricF
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.base_weight_quantizer_autograd_function import \
        BaseWeightQuantizerAutogradFunction
    import torch


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.LUT_POT_QUANTIZER],
                    identifier=QuantizerID.INFERABLE)
    class WeightsLUTPOTInferableQuantizer(WeightsLUTSymmetricInferableQuantizer):
        """
        Class for quantizing weights using lut power-of-two quantizer
        """

        def __init__(self,
                     num_bits: int,
                     lut_values: List[float],
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None,
                     lut_values_bitwidth: int = LUT_VALUES_BITWIDTH,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                lut_values: the values in the look-up table to assign the weights to
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on
                lut_values_bitwidth: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """
            # target of Weights quantization
            super(WeightsLUTPOTInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                  threshold=threshold,
                                                                  lut_values=lut_values,
                                                                  per_channel=per_channel,
                                                                  channel_axis=channel_axis,
                                                                  input_rank=input_rank,
                                                                  lut_values_bitwidth=lut_values_bitwidth,
                                                                  eps=eps)
            is_threshold_pot = np.all(np.round(np.log2(self._threshold_np.flatten())) == np.log2(self._threshold_np.flatten()))
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {threshold}'


        def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """

            if self._use_custom_impl and torch.jit.is_tracing():
                return WeightsLUTPOTF.apply(inputs,
                                            self.num_bits,
                                            self._lut_values_np,
                                            self._threshold_np,
                                            self.lut_values_bitwidth,
                                            self.eps,
                                            self.per_channel,
                                            self.channel_axis,
                                            self.input_rank
                                            )

            return super(WeightsLUTPOTInferableQuantizer, self).__call__(inputs)


    class WeightsLUTPOTF(BaseWeightQuantizerAutogradFunction):
        """
        Custom autograd function for symmetric weights quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """
        @staticmethod
        def forward(ctx,
                    input_tensor,
                    num_bits,
                    lut_values,
                    threshold,
                    lut_values_bitwidth,
                    eps,
                    per_channel,
                    channel_axis,
                    input_rank
                    ):
            return WeightsLUTSymmetricF.forward(ctx,
                                                input_tensor,
                                                num_bits,
                                                lut_values,
                                                threshold,
                                                lut_values_bitwidth,
                                                eps,
                                                per_channel,
                                                channel_axis,
                                                input_rank
                                                )

        @staticmethod
        def symbolic(g,
                     input_tensor,
                     num_bits,
                     lut_values,
                     threshold,
                     lut_values_bitwidth,
                     eps,
                     per_channel,
                     channel_axis,
                     input_rank):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                num_bits: The number of bits to represent the quantized value.
                threshold: The quantization threshold.
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            # When None is passed as channel_axis, the op has no attribute of channel_axis,
            # which creates conflict with the onnxruntime function. For this reason, if we quantize
            # per-tensor and channel_axis is None, we set it to 0.
            # per-tensor and input_rank is None, we set it to 4.
            if not per_channel and channel_axis is None:
                channel_axis = 0
            if not per_channel and input_rank is None:
                input_rank = 4

            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsLUTPOTQuantizer", input_tensor,
                        g.op('Constant', value_t=torch.tensor(lut_values, dtype=torch.float32)),
                        g.op('Constant', value_t=torch.tensor(threshold, dtype=torch.float32)),
                        num_bits_i=num_bits,
                        per_channel_i=int(per_channel),
                        channel_axis_i=channel_axis,
                        input_rank_i=input_rank,
                        lut_values_bitwidth_i=lut_values_bitwidth,
                        eps_f=eps,
                        signed_i=int(WeightsLUTPOTF.is_signed()),
                        **WeightsLUTPOTF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())


else:
    class WeightsLUTPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using WeightsLUTPOTInferableQuantizer. '
                            'Could not find torch package.')


if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.weights_lut_symmetric_inferable_quantizer import  quantize_lut_sym_weights_numpy

    # Add onnx op function to use during onnxruntime WeightsLUTPOTQuantizer op inference
    # Using this decorator the op WeightsLUTPOTQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsLUTPOTQuantizer",
             inputs=[PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={
                 "lut_values_bitwidth": PyCustomOpDef.dt_int64,
                 "eps": PyCustomOpDef.dt_float,
                 "per_channel": PyCustomOpDef.dt_int64,
                 "channel_axis": PyCustomOpDef.dt_int64,
                 "input_rank": PyCustomOpDef.dt_int64
             })
    def weight_lut_sym_ort(input_tensor: np.ndarray,
                           lut_values: np.ndarray,
                           threshold: np.ndarray,
                           **kwargs):
        return quantize_lut_sym_weights_numpy(input_tensor,
                                              lut_values,
                                              threshold,
                                              kwargs["lut_values_bitwidth"],
                                              kwargs["eps"],
                                              kwargs["per_channel"],
                                              kwargs["channel_axis"],
                                              kwargs["input_rank"])


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, FOUND_ONNXRUNTIME_EXTENSIONS, ONNX_CUSTOM_OP_DOMAIN
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.common.quant_utils import adjust_range_to_include_zero
from mct_quantizers.pytorch.onnxruntime_validations import validate_activation_params

if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizers.base_uniform_inferable_quantizer import BaseUniformInferableQuantizer
    from mct_quantizers.pytorch.quantizers.activation_inferable_quantizers.base_activation_quantizer_autograd_function \
        import \
        BaseActivationQuantizerAutogradFunction

    def quantize_uniform_activations_torch(tensor_data: torch.Tensor,
                                           range_min: float,
                                           range_max: float,
                                           n_bits: int) -> np.ndarray:
        """
       Quantizes the input tensor uniformly using PyTorch.

       Args:
           tensor_data (torch.Tensor): The input tensor to be quantized.
           range_min (float): The quantization min range.
           range_max (float): The quantization max range.
           n_bits (int): Number of bits to represent the quantized value.

       Returns:
           torch.Tensor: Uniformly quantized tensor.
       """

        # range_min = torch.tensor([range_min]).to(get_working_device())
        # range_max = torch.tensor([range_max]).to(get_working_device())

        # adjusts the quantization rage so the quantization grid include zero.
        # a, b = fix_range_to_include_zero(range_min, range_max, n_bits)

        a, b = adjust_range_to_include_zero(range_min, range_max, n_bits)

        # Compute the step size of quantized values.
        delta = (b - a) / (2 ** n_bits - 1)

        # Clip data in range
        clipped_tensor = torch.clip(tensor_data, min=a, max=b)

        # Quantize the data between min/max of quantization range.
        q = delta * torch.round((clipped_tensor - a) / delta) + a
        return q


    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    identifier=QuantizerID.INFERABLE)
    class ActivationUniformInferableQuantizer(BaseUniformInferableQuantizer):
        """
        Class for quantizing activations using an uniform quantizer
        """

        def __init__(self,
                     num_bits: int,
                     min_range: List[float],
                     max_range: List[float]
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min range for quantizing activations
                max_range: max range for quantizing activations
            """
            super(ActivationUniformInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                      min_range=min_range,
                                                                      max_range=max_range)

            assert isinstance(min_range, list), f'min_range is expected to be a list, but is of type {type(min_range)}'
            assert isinstance(max_range, list), f'max_range is expected to be a list, but is of type {type(max_range)}'

            assert len(
                min_range) == 1, f'For activation, only per-tensor quantization is supported. Thus, min_range should be ' \
                                 f'of length 1 but is {len(min_range)}'
            assert len(
                max_range) == 1, f'For activation, only per-tensor quantization is supported. Thus, max_range should be ' \
                                 f'of length 1 but is {len(max_range)}'

            # Activation is per-tensor thus we expect only a single min/max values
            self.min_range = self.min_range[0].cpu().item()
            self.max_range = self.max_range[0].cpu().item()

            self.scale = float((self.max_range-self.min_range) / ((2 ** num_bits) - 1))
            self.zero_point = int(-np.round(self.min_range / self.scale))  # zp has to be positive, and a <=0, so we multiply by -1

        def __call__(self, inputs: torch.Tensor):
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            if self._use_custom_impl and torch.jit.is_tracing():
                return ActivationUniformF.apply(inputs, self.min_range, self.max_range, self.num_bits)
            else:
                with torch.no_grad():
                    return torch.fake_quantize_per_tensor_affine(inputs,
                                                                 scale=self.scale,
                                                                 zero_point=self.zero_point,
                                                                 quant_min=self.min_quantized_domain,
                                                                 quant_max=self.max_quantized_domain)


    class ActivationUniformF(BaseActivationQuantizerAutogradFunction):
        """
        Custom autograd function for uniform activations quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, min_range, max_range, num_bits):
            """
             Forward computation function. This method performs the forward computation using
             the given quantize_uniform_activations_torch function.

             Args:
                 ctx: An object that can be used to stash information for backward function.
                 input_tensor: The input tensor to be quantized.
                 min_range: The quantization min range.
                 max_range: The quantization max range.
                 num_bits: The number of bits to represent the quantized tensor.

             Returns:
                 The quantized tensor.
             """
            return quantize_uniform_activations_torch(input_tensor, min_range, max_range, num_bits)

        @staticmethod
        def symbolic(g, input_tensor, min_range, max_range, num_bits):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                min_range: The quantization min range.
                max_range: The quantization max range.
                num_bits: The number of bits to represent the quantized value.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::ActivationUniformQuantizer", input_tensor,
                        min_range_f=min_range,
                        max_range_f=max_range,
                        num_bits_i=num_bits,
                        **ActivationUniformF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())


else:
    class ActivationUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using ActivationUniformInferableQuantizer. '
                            'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef

    def quantize_uniform_activations_numpy(tensor_data: np.ndarray,
                                           range_min: float,
                                           range_max: float,
                                           n_bits: int) -> np.ndarray:
        """
        Quantize a tensor according to given range (min, max) and number of bits.

        Args:
            tensor_data: Tensor values to quantize.
            range_min: minimum bound of the range for quantization (or array of min values per channel).
            range_max: maximum bound of the range for quantization (or array of max values per channel).
            n_bits: Number of bits to quantize the tensor.

        Returns:
            Quantized data.
        """

        validate_activation_params(input_tensor=tensor_data,
                               min_range=range_min,
                               max_range=range_max)

        # adjusts the quantization rage so the quantization grid include zero.
        a, b = adjust_range_to_include_zero(range_min, range_max, n_bits)

        # Compute the step size of quantized values.
        delta = (b - a) / (2 ** n_bits - 1)

        # Clip data in range
        clipped_tensor = np.clip(tensor_data, a_min=a, a_max=b)

        # Quantize the data between min/max of quantization range.
        q = delta * np.round((clipped_tensor - a) / delta) + a
        return q


    # Add onnx op function to use during onnxruntime ActivationUniformQuantizer op inference
    # Using this decorator the op ActivationUniformQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::ActivationUniformQuantizer",
             inputs=[PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={"num_bits": PyCustomOpDef.dt_int64,
                    "min_range": PyCustomOpDef.dt_float,
                    "max_range": PyCustomOpDef.dt_float}
             )
    def activation_uniform_ort(input_tensor: np.ndarray,
                               **kwargs):
        return quantize_uniform_activations_numpy(input_tensor,
                                                  kwargs["min_range"],
                                                  kwargs["max_range"],
                                                  kwargs["num_bits"])



import importlib
import unittest

from mct_quantizers.common.constants import TORCH, TENSORFLOW

found_tf = importlib.util.find_spec(TENSORFLOW) is not None
found_pytorch = importlib.util.find_spec(TORCH)

if __name__ == '__main__':
    # -----------------  Load all the test cases
    suiteList = []
    test_loader = unittest.TestLoader()
    MCTQuantizersSuite = unittest.TestSuite()

    # Add TF tests only if tensorflow is installed
    if found_tf:
        package_tests = test_loader.discover("tests.keras_tests", pattern="test_*.py")
        MCTQuantizersSuite.addTests(package_tests)

    # Add Pytorch tests only if Pytorch is installed
    if found_pytorch:
        package_tests = test_loader.discover("tests.pytorch_tests", pattern="test_*.py")
        MCTQuantizersSuite.addTests(package_tests)

    unittest.TextTestRunner(verbosity=0).run(MCTQuantizersSuite)


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, List, Any, Tuple, Union
from packaging import version
import numpy as np

from mct_quantizers.common.base_inferable_quantizer import BaseInferableQuantizer
from mct_quantizers.common.constants import FOUND_TF, WEIGHTS_QUANTIZERS, STEPS, WEIGHTS_VALUES, IS_INPUT_AS_LIST, \
    OP_CALL_ARGS, OP_CALL_KWARGS, LAYER, TRAINING, MCTQ_VERSION, POSITIONAL_WEIGHT, QUANTIZED_POSITIONAL_WEIGHT
from mct_quantizers.logger import Logger
from mct_quantizers.common.get_all_subclasses import get_all_subclasses
from mct_quantizers import __version__ as mctq_version

if FOUND_TF:
    import tensorflow as tf
    from tensorflow.python.util import tf_inspect
    from tensorflow.python.keras.utils.control_flow_util import smart_cond

    from mct_quantizers.keras.quantizers import BaseKerasInferableQuantizer

    keras = tf.keras

    def _make_quantizer_fn(quantizer, x, training):
        """Use currying to return True/False specialized fns to the cond."""

        def quantizer_fn():
            return quantizer(x, training)

        return quantizer_fn


    def _weight_name(name: str) -> str:
        """Extracts the weight name from the full TensorFlow variable name.

        For example, returns 'kernel' for 'dense_2/kernel:0'.

        Args:
          name: TensorFlow variable name.

        Returns:
          Extracted weight name.
        """
        return name.split(':')[0].split('/')[-1]


    def _serialize_object(obj: Union[tf.Tensor, np.ndarray]) -> Dict:
        """
        A serialization function to replace keras.utils.serialize_keras_object for TF version < 2.13,
        which fails for objects of type tf.Tensor and np.ndarray, and mimic that function's operation
        in TF version >= 2.13.
        May be deleted when TF 2.12 is no longer supported.

        Args:
          obj: Object to be serialized.

        Returns:
            A dictionary with the object's config.
        """
        if isinstance(obj, tf.Tensor):
            return {'class_name': '__tensor__',
                    'config': {'value': obj.numpy().tolist(),
                               'dtype': obj.dtype.name}}
        elif isinstance(obj, np.ndarray):
            return {'class_name': '__numpy__',
                    'config': {'value': obj.tolist(),
                               'dtype': obj.dtype.name}}
        else:
            Logger.error(f'_serialize_object only accepts tf.Tensor or np.ndarray but got type {type(obj)}.')


    class KerasQuantizationWrapper(tf.keras.layers.Wrapper):
        def __init__(self,
                     layer: tf.keras.layers.Layer,
                     weights_quantizers: Dict[Union[int, str], BaseInferableQuantizer],
                     weight_values: Dict[int, Union[np.ndarray, tf.Tensor]] = None,
                     op_call_args: List = None,
                     op_call_kwargs: Dict[str, Any] = None,
                     is_inputs_as_list: bool = False,
                     **kwargs):
            """
            The KerasQuantizationWrapper takes a keras layer and quantization information and creates
            a quantized layer. The quantization information includes a quantizer per layer attribute for
            a keras layer that contains weight attributes (e.g. Conv2D, BatchNormalization, etc.). For
            layers that get constants (e.g. tf.add(KerasTensor, tf.constant), the quantization information
            also includes a weight values per attribute, the function call args & kwargs and a boolean for
            whether the layer\function accepts the inputs as a list (e.g. tf.concat or layers.Add). Note
            that for a layer with constants, the constants are referred to as "positional weights" whose
            attributes are integers representing the input index in the function\layer's inputs.

            Args:
                layer: A keras layer.
                weights_quantizers: A dictionary between a weight's name or position to its quantizer.
                weight_values: A dictionary between a weight's position to its value.
                op_call_args: A list containing the layer's call arguments.
                op_call_kwargs: A dictionary containing the layer's call keyword arguments.
                is_inputs_as_list: A boolean indicating the layer accepts the input tensors as a list.

            Examples:

                Creating a quantized Conv2D (kernel only):

                >>> import mct_quantizers as mctq
                >>> import tensorflow as tf

                >>> attr_quant_dict = {'kernel': mctq.keras.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False)}
                >>> QuantizedConv2D = mctq.KerasQuantizationWrapper(tf.keras.layers.Conv2D(3,3), attr_quant_dict)

                creating a quantized function with a constant: tf.subtract(tf.constant, KerasTensor)

                >>> attr_quant_dict = {0: mctq.keras.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False)}
                >>> attr_values = {0: tf.constant([1, 2, 3], dtype=tf.float32)}
                >>> QuantizedSub = mctq.KerasQuantizationWrapper(TFOpLambda(tf.subtract), attr_quant_dict, attr_values)

                creating a quantized function with a constant and arguments: tf.matmul(KerasTensor, tf.constant, transpose_b=True)
                >>> attr_quant_dict = {1: mctq.keras.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False)}
                >>> attr_values = {1: tf.constant([[1,2,3], [4, 5, 6]], dtype=tf.float32)}
                >>> QuantizedMatmul = mctq.KerasQuantizationWrapper(TFOpLambda(tf.matmul), attr_quant_dict,
                >>>                                                 attr_values, op_call_kwargs={'transpose_b', True})

            """
            super(KerasQuantizationWrapper, self).__init__(layer, **kwargs)
            self._track_trackable(layer, name='layer')
            # making sure the attribute name is converted to the actual attribute field name in the layer.
            self.weights_quantizers = {_weight_name(k) if isinstance(k ,str) else k: v
                                       for k, v in weights_quantizers.items()}
            # Initialize positional weights:
            self.weight_values = weight_values if weight_values is not None else dict()
            for pos, weight_val in self.weight_values.items():
                if not isinstance(weight_val, (np.ndarray, tf.Tensor)):
                    Logger.error(f'Positional weight at position {pos} should be either an ndarray or a tf.Tensor,'
                                 f'but type is {type(weight_val)}')
            if version.parse(tf.__version__) < version.parse("2.13"):
                # Convert all values to tensors because keras.utils.serialize_keras_object fails for numpy array
                # before version 2.13. (TODO: remove this if-else when not supporting TF 2.12 and below)
                self.serialize_fn = _serialize_object
            else:
                self.serialize_fn = keras.utils.serialize_keras_object

            # Initialize functional layer arguments. For examples, see the class description.
            self.op_call_args = [] if op_call_args is None else op_call_args
            self.op_call_kwargs = {} if op_call_kwargs is None else op_call_kwargs
            self.is_inputs_as_list = is_inputs_as_list

            # Sanity checks
            # 1. If there are no weight_values: verify all weight_quantizers are strings
            # 2. If there are weight_values: verify all weight_quantizers and weight_values keys
            #    are integers, and that they match.
            # 3. A layer with both weights as attributes and positional weights is not supported.
            if len(self.weight_values) == 0:
                # expecting weights_quantizers keys to be all strings
                if not all([isinstance(w, str) for w in self.weights_quantizers]):
                    Logger.error('"weights_quantizers" keys should be all strings')
                self.is_str_attr = True
            else:
                # expecting both weights_quantizers and weight_values keys to be all integers.
                if not all([isinstance(w, int) for w in self.weight_values]):
                    Logger.error('All "weight_values" keys should be integers.')
                if not all([a == b for a, b in zip(weights_quantizers, weight_values)]):
                    Logger.error('Mismatch between "weights_quantizers" and "weight_values" keys')
                self.is_str_attr = False

            if not all([isinstance(w, (int, str)) for w in weights_quantizers]):
                Logger.error('All "weight_values" keys should be either strings ot integers')

            self._mctq_version = mctq_version

        @property
        def is_weights_quantization(self) -> bool:
            """
            This function check weights quantizer exists in wrapper.

            Returns: a boolean if weights quantizer exists.

            """
            return self.num_weights_quantizers > 0

        @property
        def num_weights_quantizers(self) -> int:
            """
            Returns: number of weights quantizers.
            """
            return len(self.weights_quantizers)

        def get_config(self):
            """
            Returns: Configuration of KerasQuantizationWrapper.

            """
            base_config = super(KerasQuantizationWrapper, self).get_config()
            config = {WEIGHTS_QUANTIZERS: {k: keras.utils.serialize_keras_object(v) for k, v in self.weights_quantizers.items()}}
            # Only create the wrapper attributes that handle positional weights if they exist, so the wrapper is forward
            # compatible with older MCTQ versions (at least until MCT will start quantizing positional weights)
            if len(self.weight_values) > 0:
                config[WEIGHTS_VALUES] = {k: self.serialize_fn(v) for k, v in self.weight_values.items()}
                config[OP_CALL_ARGS] = self.op_call_args
                config[OP_CALL_KWARGS] = self.op_call_kwargs
                config[IS_INPUT_AS_LIST] = self.is_inputs_as_list
            return_config = {**base_config, **config}
            return_config[MCTQ_VERSION] = self._mctq_version

            return return_config

        @property
        def mctq_version(self):
            return self._mctq_version

        def _set_weights_vars(self, is_training: bool = True):
            """
            This function sets weights quantizers vars to the layer.

            Args:
                is_training: Flag to indicate whether training or not.

            Returns: None
            """
            self._weights_vars = []
            for name, quantizer in self.weights_quantizers.items():
                if isinstance(name, str):
                    weight = getattr(self.layer, name)
                    _name = _weight_name(weight.name)
                    if is_training and not any([weight is w for w in self._trainable_weights]):
                        self._trainable_weights.append(weight)
                    elif not is_training and any([weight is w for w in self._non_trainable_weights]):
                        self._non_trainable_weights.append(weight)
                elif isinstance(name, int):
                    weight_value = self.weight_values[name]
                    _name = None
                    weight = self.add_weight(name=f'{POSITIONAL_WEIGHT}_{name}',
                                             shape=weight_value.shape,
                                             initializer=tf.keras.initializers.Constant(weight_value),
                                             trainable=False)
                    setattr(self, f'{POSITIONAL_WEIGHT}_{name}', weight)
                else:
                    Logger.error(f'A weight name ({name}) should be either "str" or "int", but has type {type(name)}')

                quantizer.initialize_quantization(weight.shape, _name if is_training else None,
                                                  self)

                # Add weight to wrapper weight lists (rather than the layer weight lists), because it will be deleted
                # from the layer's lists after the first call.
                self._weights_vars.append((name, weight, quantizer))

        @classmethod
        def from_config(cls, config):
            """

            Args:
                config(dict): dictionary  of  KerasQuantizationWrapper Configuration.

            Returns: A KerasQuantizationWrapper.

            """
            numpy_deserialization = lambda **_config: tf.constant(**_config).numpy()
            tensor_deserialization = lambda **_config: tf.constant(**_config)
            # When reading weight quantizers keys, which may be either a string with attribute name or an integer, the
            # key is always a string, so this function checks whether it was a string or integer before serialization.
            maybe_int = lambda x: int(x) if x.isdigit() else x

            config = config.copy()
            qi_inferable_custom_objects = {subclass.__name__: subclass for subclass in
                                           get_all_subclasses(BaseKerasInferableQuantizer)}
            with keras.utils.custom_object_scope(qi_inferable_custom_objects):
                weights_quantizers = {maybe_int(k): keras.utils.deserialize_keras_object(v, module_objects=globals())
                                      for k, v in config.pop(WEIGHTS_QUANTIZERS).items()}

            # read weights_values in this scope so deserialize_keras_object knows how to interpret the serialization.
            with keras.utils.custom_object_scope({'__numpy__': numpy_deserialization,
                                                  '__tensor__': tensor_deserialization}):
                weights_values = {int(k): keras.utils.deserialize_keras_object(v)
                                  for k, v in config.pop(WEIGHTS_VALUES, {}).items()}

            layer = tf.keras.layers.deserialize(config.pop(LAYER))

            op_call_args = config.pop(OP_CALL_ARGS, [])
            op_call_kwargs = config.pop(OP_CALL_KWARGS, {})
            is_inputs_as_list = config.pop(IS_INPUT_AS_LIST, False)

            v = config.pop(MCTQ_VERSION, None)

            obj = cls(layer=layer, weights_quantizers=weights_quantizers, weight_values=weights_values,
                      op_call_args=op_call_args, op_call_kwargs=op_call_kwargs,
                      is_inputs_as_list=is_inputs_as_list, **config)
            obj._mctq_version = mctq_version if v is None else v

            return obj

        def build(self, input_shape):
            """
            KerasQuantization Wrapper build function.
            Args:
                input_shape: the layer input shape.

            Returns: None

            """
            super(KerasQuantizationWrapper, self).build(input_shape)

            self.optimizer_step = self.add_weight(
                STEPS,
                initializer=tf.keras.initializers.Constant(-1),
                dtype=tf.dtypes.int32,
                trainable=False)

            self._set_weights_vars()

        def set_quantize_weights(self, quantized_weights: dict):
            """
            This function update layer weights after quantization.

            Args:
                quantized_weights: a dict of weight to update.

            Returns: None

            """
            for weight_attr in self.weights_quantizers:
                weight = quantized_weights.get(weight_attr)
                if isinstance(weight_attr, str):
                    # weight attribute is a string --> weight exists as attribute in layer so
                    # override is with quantized weight.
                    current_weight = getattr(self.layer, weight_attr)
                    setattr(self.layer, weight_attr, weight)
                elif isinstance(weight_attr, int):
                    # weight attribute is an integer --> weight doesn't exist as attribute in layer
                    # so create an attribute in wrapper for the quantized weight.
                    current_weight = getattr(self, f'{POSITIONAL_WEIGHT}_{weight_attr}')
                    setattr(self, f'{QUANTIZED_POSITIONAL_WEIGHT}_{weight_attr}', weight)
                else:
                    Logger.error(f'weight attribute should be either a string or an integer, but it is {type(weight_attr)}')  # pragma: no cover

                if current_weight.shape != weight.shape:
                    Logger.error(
                        f"Existing layer weight shape {current_weight.shape} is incompatible with provided weight "
                        f"shape {weight.shape}")  # pragma: no cover

        def call(self, inputs, training=None, **kwargs):
            """
            KerasQuantizationWrapper call functions.
            Args:
                inputs: Input tensors to specified layer.
                training: a boolean stating if layer is in training mode.
                **kwargs:

            Returns: tensors that simulate a quantized layer.

            """
            if training is not None and not isinstance(training, bool):
                Logger.critical(f"Argument type mismatch: expected 'training' argument to be of type 'boolean' "
                                f"but got {type(training)}")

            if training is None:
                training = tf.keras.backend.learning_phase()

            # Quantize all weights, and replace them in the underlying layer.
            quantized_weights = {}
            for name, unquantized_weight, quantizer in self._weights_vars:

                weights_quantizer_args_spec = tf_inspect.getfullargspec(quantizer.__call__).args
                if TRAINING in weights_quantizer_args_spec:
                    quantized_weight = smart_cond(
                        training,
                        _make_quantizer_fn(quantizer, unquantized_weight, True),
                        _make_quantizer_fn(quantizer, unquantized_weight, False))
                    quantized_weights.update({name: quantized_weight})
                else:
                    # Keras weights inferable quantizer.
                    quantized_weight = quantizer(unquantized_weight)
                    quantized_weights.update({name: quantized_weight})

            self.set_quantize_weights(quantized_weights)

            if self.is_str_attr:
                args_spec = tf_inspect.getfullargspec(self.layer.call).args
                if TRAINING in args_spec:
                    kwargs.update({'training': training})
                outputs = self.layer.call(inputs, **kwargs)
            else:
                _inputs = inputs if isinstance(inputs, list) else [inputs]
                weight_positions = [w[0] for w in self._weights_vars]
                for pos in sorted(weight_positions):
                    _inputs.insert(pos, getattr(self, f'{QUANTIZED_POSITIONAL_WEIGHT}_{pos}'))

                if self.is_inputs_as_list:
                    outputs = self.layer.call(_inputs, *self.op_call_args, **self.op_call_kwargs)
                else:
                    outputs = self.layer.call(*(_inputs + self.op_call_args), **self.op_call_kwargs)

            return outputs

        def get_weights_vars(self) -> List[Tuple[str, Any, BaseInferableQuantizer]]:
            """
            A getter of the layer's weights variables.

            Returns:
                List pf tuples of the wrapped layer's weights variables with weight name, values and assigned quantizer.

            """

            return self._weights_vars

        def get_quantized_weights(self) -> Dict[str, tf.Tensor]:
            """

            Returns: A dictionary of weights attributes to quantized weights.

            """
            quantized_weights = {}
            weights_var = self.get_weights_vars()
            for name, w, quantizer in weights_var:
                quantized_weights[name] = quantizer(w)
            return quantized_weights

else:
    class KerasQuantizationWrapper:
        def __init__(self,
                     layer,
                     weights_quantizers: Dict[str, BaseInferableQuantizer],
                     weight_values: Dict = None,
                     op_call_args: List = None,
                     op_call_kwargs: Dict[str, Any] = None,
                     is_inputs_as_list: bool = False):
            """
            Keras Quantization Wrapper takes a keras layer and quantizers and infer a quantized layer.

            Args:
                layer: A keras layer.
                weights_quantizers: A dictionary between a weight's name to its quantizer.
                weight_values: A dictionary between a weight's position to its value.
                op_call_args: A list containing the layer's call arguments.
                op_call_kwargs: A dictionary containing the layer's call keyword arguments.
                is_inputs_as_list: A boolean indicating the layer accepts the input tensors as a list.
            """
            Logger.critical('Installing tensorflow is mandatory '
                            'when using KerasQuantizationWrapper. '
                            'Could not find Tensorflow package.')  # pragma: no cover



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Union, Any, Dict, Tuple, Callable

import inspect

from mct_quantizers.common.base_inferable_quantizer import BaseInferableQuantizer
from mct_quantizers.common.constants import FOUND_TORCH, LAYER, TRAINING, POSITIONAL_WEIGHT, \
    QUANTIZED_POSITIONAL_WEIGHT
from mct_quantizers.logger import Logger

if FOUND_TORCH:
    import torch
    import torch.nn as nn


    class PytorchQuantizationWrapper(nn.Module):
        def __init__(self,
                     module: Union[nn.Module, Callable],
                     weights_quantizers: Dict[Union[int, str], BaseInferableQuantizer],
                     weight_values: Dict[int, torch.Tensor] = None,
                     op_call_args: List = None,
                     op_call_kwargs: Dict[str, Any] = None,
                     is_inputs_as_list: bool = False):
            """
            The PytorchQuantizationWrapper takes a Pytorch layer and quantization information and creates
            a quantized layer. The quantization information includes a quantizer per layer attribute for
            a Torch layer that contains weight attributes (e.g. Conv2d, BatchNorm2d, etc.). For
            layers that get constants (e.g. torch.add(Tensor, constant), the quantization information
            also includes a weight values per attribute, the function call args & kwargs and a boolean for
            whether the layer\function accepts the inputs as a list (e.g. torch.cat). Note that for a layer
            with constants, the constants are referred to as "positional weights" whose attributes are integers
            representing the input index in the function\layer's inputs.

            Args:
                module: A pytorch module or as function.
                weights_quantizers: A dictionary between a weight's name or position to its quantizer.
                weight_values: A dictionary between a weight's position to its value.
                op_call_args: A list containing the layer's call arguments.
                op_call_kwargs: A dictionary containing the layer's call keyword arguments.
                is_inputs_as_list: A boolean indicating the layer accepts the input tensors as a list.

            Examples:

                Creating a quantized Conv2d (weight only):

                >>> import mct_quantizers as mctq
                >>> import torch

                >>> attr_quant_dict = {'weight': mctq.pytorch.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False)}
                >>> QuantizedConv2D = mctq.PytorchQuantizationWrapper(torch.nn.Conv2d(3, 3, 3), attr_quant_dict)

                creating a quantized function with a constant: torch.sub(constant, Tensor)

                >>> attr_quant_dict = {0: mctq.pytorch.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False)}
                >>> attr_values = {0: torch.Tensor([1, 2, 3], dtype=torch.float32)}
                >>> QuantizedSub = mctq.PytorchQuantizationWrapper(torch.sub), attr_quant_dict, attr_values)

                creating a quantized function with constants and arguments: tf.cat([constant#1, Tensor, constant#2], dim=1)
                >>> attr_quant_dict = {0: mctq.pytorch.quantizers.WeightsPOTInferableQuantizer(4, [2.0], False),
                >>>                    2: mctq.pytorch.quantizers.WeightsPOTInferableQuantizer(4, [1.0], False)}
                >>> attr_values = {0: torch.Tensor([[1,2,3], [4, 5, 6]], dtype=torch.float32),
                >>>                2: torch.Tensor([[4,5,6], [4, 5, 6]], dtype=torch.float32)}
                >>> QuantizedConcat = mctq.PytorchQuantizationWrapper(torch.cat, attr_quant_dict, attr_values,
                >>>                                                   op_call_kwargs={'dim', 1})

            """
            super().__init__()
            if isinstance(module, nn.Module):
                self.add_module(LAYER, module)
            else:
                # Functional layers
                setattr(self, LAYER, module)

            self.weights_quantizers = weights_quantizers
            # Initialize positional weights:
            self.weight_values = weight_values if weight_values is not None else dict()
            for pos, weight_val in self.weight_values.items():
                if not isinstance(weight_val, torch.Tensor):
                    Logger.error(f'Positional weight at position {pos} should be a torch.Tensor, '
                                 f'but type is {type(weight_val)}.')

            # Initialize functional module arguments. For examples, see the class description.
            self.op_call_args = [] if op_call_args is None else op_call_args
            self.op_call_kwargs = {} if op_call_kwargs is None else op_call_kwargs
            self.is_inputs_as_list = is_inputs_as_list

            # Sanity checks:
            # 1. If there are no weight_values: verify all weight_quantizers are strings
            # 2. If there are weight_values: verify all weight_quantizers and weight_values keys
            #    are integers, and that they match.
            # 3. A module with both weights as attributes and positional weights is not supported.
            if len(self.weight_values) == 0:
                # expecting weights_quantizers keys to be all strings.
                if not all([isinstance(w, str) for w in self.weights_quantizers]):
                    Logger.error('"weights_quantizers" keys should be all strings')
                self.is_str_attr = True
            else:
                # expecting both weights_quantizers and weight_values keys to be all integers.
                if not all([isinstance(w, int) for w in self.weight_values]):
                    Logger.error('All "weight_values" keys should be integers')
                if not all([a == b for a, b in zip(weights_quantizers, weight_values)]):
                    Logger.error('Mismatch between "weights_quantizers" and "weight_values" keys')
                self.is_str_attr = False

            self._set_weights_vars(True)

        @property
        def is_weights_quantization(self) -> bool:
            """
            This function check weights quantizer exists in wrapper.

            Returns: a boolean if weights quantizer exists.

            """
            return self.num_weights_quantizers > 0

        @property
        def num_weights_quantizers(self) -> int:
            """
            Returns: number of weights quantizers.
            """
            return len(self.weights_quantizers)

        def convert_to_inferable_quantizers(self):
            """
            Convert the wrapper quantizers with inferable quantizers.

            """
            # Weight quantizers
            if self.is_weights_quantization:
                inferable_weight_quantizers = {}
                for name, quantizer in self.weights_quantizers.items():
                    if hasattr(quantizer, 'convert2inferable') and callable(quantizer.convert2inferable):
                        inferable_weight_quantizers.update({name: quantizer.convert2inferable()})
                self.weights_quantizers = inferable_weight_quantizers
                self._set_weights_vars(False)

        def _set_weights_vars(self, is_training: bool = True):
            """
            Initialize learnable weights as parameters in the wrapper, and their quantizers.

            Args:
                is_training: Whether working with InferableQuantizers or not. If so, do not register weight as parameter.

            """
            self._weights_vars = []

            # Init weights quantizers
            for name, quantizer in self.weights_quantizers.items():
                if self.is_str_attr:
                    if is_training:
                        weight = getattr(self.layer, name).detach()
                        delattr(self.layer, name)
                        setattr(self.layer, name, weight)
                        self.register_parameter(name, torch.nn.Parameter(weight, requires_grad=True))
                    else:
                        weight = getattr(self, name).detach()
                        delattr(self.layer, name)
                        setattr(self.layer, name, weight)
                    weight_var = getattr(self, name)
                else:
                    weight = self.weight_values[name]
                    self.register_parameter(f'{POSITIONAL_WEIGHT}_{name}',
                                            torch.nn.Parameter(weight, requires_grad=False))
                    setattr(self, f'{QUANTIZED_POSITIONAL_WEIGHT}_{name}', weight)
                    weight_var = getattr(self, f'{POSITIONAL_WEIGHT}_{name}')

                quantizer.initialize_quantization(weight.shape, name, self)
                self._weights_vars.append((name, weight_var, quantizer))

        def set_quantize_weights(self, quantized_weights: dict):
            """
            This function updates layer weights after quantization.

            Args:
                quantized_weights: a dict of weight to update.

            Returns: None

            """
            for weight_attr in self.weights_quantizers:
                weight = quantized_weights.get(weight_attr)
                if self.is_str_attr:
                    setattr(self.layer, weight_attr, weight)
                else:
                    setattr(self, f'{QUANTIZED_POSITIONAL_WEIGHT}_{weight_attr}', weight)

        def get_weights_vars(self) -> List[Tuple[str, Any, BaseInferableQuantizer]]:
            """
            A getter of the layer's weights variables.

            Returns:
                List pf tuples of the wrapped layer's weights variables with weight name, values and assigned quantizer.

            """

            return self._weights_vars

        def forward(self,
                    *args: List[Any],
                    **kwargs: Dict[str, Any]) -> Union[torch.Tensor, List[torch.Tensor]]:
            """
            PytorchQuantizationWrapper forward functions.
            Args:
                args: arguments to pass to internal layer.
                kwargs: key-word dictionary to pass to the internal layer.

            Returns: a tensor that simulates a quantized layer.

            """

            # ----------------------------------
            # Quantize all weights, and replace them in the underlying layer.
            # ----------------------------------
            if self.is_weights_quantization:

                quantized_weights = {}
                for name, unquantized_weight, quantizer in self._weights_vars:
                    s = inspect.signature(quantizer.__call__)
                    if TRAINING in s.parameters.keys():
                        quantized_weight = quantizer(unquantized_weight, self.training)
                    else:
                        quantized_weight = quantizer(unquantized_weight)

                    quantized_weights.update({name: quantized_weight})

                self.set_quantize_weights(quantized_weights)

            if not self.is_str_attr:
                # Positional weights need to be inserted in the wrapper input list according to their (key) position.
                args = list(args)
                weight_positions = [w[0] for w in self._weights_vars]
                for pos in sorted(weight_positions):
                    args.insert(pos, getattr(self, f'{QUANTIZED_POSITIONAL_WEIGHT}_{pos}'))

            _kwargs = {**self.op_call_kwargs, **kwargs}
            # ----------------------------------
            # Layer operation
            # ----------------------------------
            if self.is_inputs_as_list:
                outputs = self.layer(args, *self.op_call_args, **_kwargs)
            else:
                outputs = self.layer(*args, *self.op_call_args, **_kwargs)

            return outputs

        def get_quantized_weights(self) -> Dict[str, torch.Tensor]:
            """

            Returns: A dictionary of weights attributes to quantized weights.

            """
            quantized_weights = {}
            weights_var = self.get_weights_vars()
            for name, w, quantizer in weights_var:
                quantized_weights[name] = quantizer(w)
            return quantized_weights

else:
    class PytorchQuantizationWrapper:
        def __init__(self,
                     layer,
                     weight_quantizers: Dict[str, BaseInferableQuantizer],
                     weight_values: Dict = None,
                     op_call_args: List = None,
                     op_call_kwargs: Dict[str, Any] = None,
                     is_inputs_as_list: bool = False):
            """
            Pytorch Quantization Wrapper takes a pytorch module and quantizers and infer a quantized layer.

            Args:
                layer: A pytorch module.
                weight_quantizers: A dictionary between a weight's name to its quantizer.
                weight_values: A dictionary between a weight's position to its value.
                op_call_args: A list containing the layer's call arguments.
                op_call_kwargs: A dictionary containing the layer's call keyword arguments.
                is_inputs_as_list: A boolean indicating the layer accepts the input tensors as a list.
            """
            Logger.critical('Installing Pytorch is mandatory '
                            'when using PytorchQuantizationWrapper. '
                            'Could not find torch package.')  # pragma: no cover



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from mct_quantizers.common.base_inferable_quantizer import BaseInferableQuantizer
from mct_quantizers.common.constants import FOUND_TORCH, ACTIVATION_HOLDER_QUANTIZER
from mct_quantizers.logger import Logger

if FOUND_TORCH:
    import torch

    class PytorchActivationQuantizationHolder(torch.nn.Module):
        """
        Pytorch module to hold an activation quantizer and quantize during inference.
        """
        def __init__(self,
                     activation_holder_quantizer: BaseInferableQuantizer,
                     **kwargs):
            """

            Args:
                activation_holder_quantizer: Quantizer to use during inference.
                **kwargs: Key-word arguments for the base layer
            """

            super(PytorchActivationQuantizationHolder, self).__init__(**kwargs)
            self.activation_holder_quantizer = activation_holder_quantizer
            self.activation_holder_quantizer.initialize_quantization(None,
                                                                     ACTIVATION_HOLDER_QUANTIZER + "_out",
                                                                     self)

        def forward(self, inputs):
            """
            Quantizes the input tensor using the activation quantizer of class PytorchActivationQuantizationHolder.

            Args:
                inputs: Input tensors to quantize with the activation quantizer.

            Returns: Output of the activation quantizer (quantized input tensor).

            """
            return self.activation_holder_quantizer(inputs)

        def convert_to_inferable_quantizers(self):
            """
            Convert a layer's quantizer to an inferable quantizer.
            Returns:
                None
            """
            if hasattr(self.activation_holder_quantizer, 'convert2inferable') and callable(
                    self.activation_holder_quantizer.convert2inferable):  # pragma: no cover
                self.activation_holder_quantizer = self.activation_holder_quantizer.convert2inferable()

else:
    class PytorchActivationQuantizationHolder:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.critical('Installing Pytorch is mandatory '
                            'when using PytorchActivationQuantizationHolder. '
                            'Could not find the torch package.')  # pragma: no cover


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import numpy as np


def adjust_range_to_include_zero(range_min: np.ndarray,
                                 range_max: np.ndarray,
                                 n_bits: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Adjusting the quantization range to include representation of 0.0 in the quantization grid.
    For per_channel quantization range_min\range_max should be tensors in the specific shape that allows
    quantization along the channel_axis.

    Args:
        range_min: min bound of the quantization range (before adjustment).
        range_max: max bound of the quantization range (before adjustment).
        n_bits: Number of bits to quantize the tensor.

    Returns: adjusted quantization range
    """
    scale = (range_max - range_min) / (2 ** n_bits - 1)
    min_range_adj = scale * np.round(range_min / scale)
    max_range_adj = range_max - range_min + min_range_adj

    min_positive = range_min > 0
    max_negative = range_max < 0
    mid_range = np.logical_and(np.logical_not(min_positive), np.logical_not(max_negative))

    min_range_adj = min_range_adj * mid_range + max_negative * range_min
    max_range_adj = max_range_adj * mid_range + min_positive * range_max

    # Make sure min_range_adj < 0 and max_range_adj > 0 to avoid small numeric error
    min_range_adj = np.minimum(min_range_adj, 0)
    max_range_adj = np.maximum(max_range_adj, 0)

    return min_range_adj, max_range_adj


def lut_quantizer_np(tensor_data: np.ndarray,
                     lut_values: np.ndarray,
                     signed: bool,
                     threshold: np.ndarray,
                     lut_values_bitwidth: int,
                     eps: float,
                     per_channel: bool,
                     channel_axis: int=None,
                     input_rank: int=None
                     ) -> np.ndarray:
    """
    Quantize a tensor using a non-uniform quantization based on the pre-defined values.
    """
    if per_channel:
        threshold_target_shape = [1] * input_rank
        threshold_target_shape[channel_axis] = -1
        threshold = np.reshape(threshold, threshold_target_shape)

    tensor = int_quantization_with_threshold(tensor_data,
                                             n_bits=lut_values_bitwidth,
                                             signed=signed,
                                             threshold=threshold,
                                             eps=eps)
    tensor = np.expand_dims(tensor, axis=-1)

    expanded_lut_values = lut_values.reshape([*[1 for _ in range(len(tensor.shape) - 1)], -1])
    lut_values_assignments = np.argmin(np.abs(tensor - expanded_lut_values), axis=-1)
    centers = lut_values.flatten()[lut_values_assignments]

    quant_tensor = (centers / (2 ** (lut_values_bitwidth - int(signed)))) * threshold

    return quant_tensor


def int_quantization_with_threshold(data: np.ndarray,
                                    n_bits: int,
                                    signed: bool,
                                    threshold: np.ndarray,
                                    eps: float) -> np.ndarray:
    """
    Divides data by threshold and quantize it to integers in the quantization range.
    """

    if signed:
        clip_max = 2 ** (n_bits - 1) - 1
        clip_min = -2 ** (n_bits - 1)
    else:
        clip_max = 2 ** n_bits - 1
        clip_min = 0

    return np.clip((data / (threshold + eps)) * (2 ** (n_bits - int(signed))),
                   clip_min, clip_max)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import importlib.util


## Frameworks

TENSORFLOW = 'tensorflow'
TORCH = 'torch'
ONNX = 'onnx'
ONNXRUNTIME = 'onnxruntime'
ONNXRUNTIME_EXTENSIONS = 'onnxruntime_extensions'

FOUND_TF = importlib.util.find_spec(TENSORFLOW) is not None
FOUND_TORCH = importlib.util.find_spec(TORCH) is not None
FOUND_ONNX = importlib.util.find_spec(ONNX) is not None
FOUND_ONNXRUNTIME = importlib.util.find_spec(ONNXRUNTIME) is not None
FOUND_ONNXRUNTIME_EXTENSIONS = importlib.util.find_spec(ONNXRUNTIME_EXTENSIONS) is not None


## Quantization properties
IS_WEIGHTS = "is_weights"
IS_ACTIVATIONS = "is_activations"
WEIGHTS_QUANTIZERS = "weights_quantizer"
WEIGHTS_VALUES = "weights_value"
OP_CALL_ARGS = 'op_call_args'
OP_CALL_KWARGS = 'op_call_kwargs'
IS_INPUT_AS_LIST = 'is_inputs_as_list'
WEIGHTS_QUANTIZATION_METHOD = 'weights_quantization_method'
WEIGHTS_N_BITS = 'weights_n_bits'
WEIGHTS_QUANTIZATION_PARAMS = 'weights_quantization_params'
ENABLE_WEIGHTS_QUANTIZATION = 'enable_weights_quantization'
WEIGHTS_CHANNELS_AXIS = 'weights_channels_axis'
WEIGHTS_PER_CHANNEL_THRESHOLD = 'weights_per_channel_threshold'
MIN_THRESHOLD = 'min_threshold'
ACTIVATION_QUANTIZATION_METHOD = 'activation_quantization_method'
ACTIVATION_N_BITS = 'activation_n_bits'
ACTIVATION_QUANTIZATION_PARAMS = 'activation_quantization_params'
ENABLE_ACTIVATION_QUANTIZATION = 'enable_activation_quantization'


## Quantizers

QUANTIZATION_TARGET = 'quantization_target'
QUANTIZATION_METHOD = 'quantization_method'
QUANTIZER_ID = 'identifier'

# In KerasQuantizationWrapper and PytorchQuantizationWrapper multiple quantizers are kept
ACTIVATION_QUANTIZERS = "activation_quantizers"
# In ActivationQuantizationHolder only one quantizer is used thus a new attribute name is needed
ACTIVATION_HOLDER_QUANTIZER = "activation_holder_quantizer"

# Quantizer signature parameters:
NUM_BITS = 'num_bits'
SIGNED = 'signed'
THRESHOLD = 'threshold'
PER_CHANNEL = 'per_channel'
MIN_RANGE = 'min_range'
MAX_RANGE = 'max_range'
CHANNEL_AXIS = 'channel_axis'
INPUT_RANK = 'input_rank'
LUT_VALUES = 'lut_values'


## Constant values

LAYER = "layer"
STEPS = "optimizer_step"
TRAINING = "training"
EPS = 1e-8
LUT_VALUES_BITWIDTH = 8

POSITIONAL_WEIGHT = 'positional_weight'
QUANTIZED_POSITIONAL_WEIGHT = f'quantized_{POSITIONAL_WEIGHT}'

# ONNX ops domain
ONNX_CUSTOM_OP_DOMAIN = f"mct_quantizers"


## Metadata common fields
FRAMEWORK_VERSION = 'framework_version'
PYTHON_VERSION = 'python_version'
MCTQ_VERSION = "mctq_version"
ONNX_VERSION = 'onnx_version'



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import warnings
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TF, LUT_VALUES_BITWIDTH, EPS
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.logger import Logger


if FOUND_TF:
    import tensorflow as tf
    from mct_quantizers.keras.quantizer_utils import lut_quantizer
    from mct_quantizers.keras.quantizers.base_keras_inferable_quantizer import BaseKerasInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.LUT_POT_QUANTIZER],
                    identifier=QuantizerID.INFERABLE)
    class ActivationLutPOTInferableQuantizer(BaseKerasInferableQuantizer):
        """
        Class for quantizing activations using lut power-of-two quantizer
        """

        def __init__(self,
                     num_bits: int,
                     lut_values: List[int],
                     threshold: List[float],
                     signed: bool,
                     lut_values_bitwidth: int = LUT_VALUES_BITWIDTH,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                lut_values: the values in the look-up table to assign the weights to
                threshold: threshold for quantizing activations
                signed: whether or not to use signed quantization
                lut_values_bitwidth: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """
            # Call the superclass constructor with the given parameters, along with the target of Activation
            # quantization
            super(ActivationLutPOTInferableQuantizer, self).__init__()

            assert isinstance(threshold, list), f'Expected threshold to be of type list but is {type(threshold)}'
            assert all([isinstance(x, (float, np.float32, tf.float32)) for x in
                        threshold]), f'Expected threshold list to contain float or np.float values but found ' \
                                     f'{[type(x) for x in threshold]}'

            # In activation per-channel quantization is not supported thus we expect a single threshold value.
            assert len(threshold) == 1, f'In per-tensor quantization threshold should be of ' \
                                        f'length 1 but is {len(threshold)}'

            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in threshold])
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {threshold}'

            self.threshold = threshold

            # Convert lut_values to numpy array for all assertions, and convert it back to list before saving.
            # The reason for doing so is that during deserialization (when get_config is called) the returned value
            # is a list (even if it was a numpy array during serialization) thus the expected lut values type
            # must be a list. The conversion to numpy is to make assertions more clean.
            lut_values = np.asarray(lut_values)

            assert len(np.unique(lut_values)) <= 2 ** num_bits, \
                f'Expected num of lut values to be less or equal than {2 ** num_bits} ' \
                f'but got {len(lut_values)}'

            assert not np.any(lut_values - lut_values.astype(int)), f'Expected lut values to be integers'

            if signed:
                assert np.all((-1 * (2 ** (lut_values_bitwidth - int(signed))) <= lut_values) &
                              (lut_values <= (2 ** (lut_values_bitwidth - int(signed)) - 1))), \
                    f'Expected lut values in the quantization range'
            else:
                assert np.all(lut_values <= (2 ** lut_values_bitwidth)), \
                    f'Expected lut values in the quantization range'

            # num_bits must be less than lut_values_bitwidth
            assert num_bits <= lut_values_bitwidth, f'Look-Up-Table bit configuration has {num_bits} bits. It must be ' \
                                                  f'less then {lut_values_bitwidth}'
            if num_bits == lut_values_bitwidth:
                warnings.warn("Num of bits equal to multiplier n bits, Please be aware LUT quantizier may be "
                              "inefficient in that case, consider using SymmetricInferableQuantizer instead")

            # If unsigned activation quantization, all lut_values must have the same sign
            if not signed:
                assert np.all(lut_values >= 0), f'Expected unsigned lut values in unsigned activation ' \
                                                     f'quantization '

            self.num_bits = num_bits
            # Save as a numpy array to avoid conversion during inference
            self._lut_values_as_np = lut_values
            # Save as a list for serialization purposes
            self.lut_values = lut_values.tolist()
            self.signed = signed
            self.lut_values_bitwidth = lut_values_bitwidth
            self.eps = eps

        def __call__(self, inputs: tf.Tensor) -> tf.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            assert inputs.dtype == tf.float32, f'Input tensor was expected to be a float tensor but is of type ' \
                                               f'{inputs.dtype}'

            return lut_quantizer(inputs,
                                 lut_values=self._lut_values_as_np.astype(np.float32),
                                 signed=self.signed,
                                 # In activation per-channel quantization is not supported
                                 # thus we expect a single threshold value. Assertion is made in init.
                                 threshold=self.threshold[0],
                                 lut_values_bitwidth=self.lut_values_bitwidth,
                                 eps=self.eps)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'lut_values', 'threshold', 'signed',
                'lut_values_bitwidth', 'eps'
            """
            return {'num_bits': self.num_bits,
                    'lut_values': self.lut_values,
                    'threshold': self.threshold,
                    'signed': self.signed,
                    'lut_values_bitwidth': self.lut_values_bitwidth,
                    'eps': self.eps}

        @classmethod
        def from_config(cls, config):
            """
            Return an object with config
            Args:
                config(dict): dictionary of object configuration
            Returns: An object created with config
            """
            return cls(config.get('num_bits'),
                       config.get('lut_values'),
                       config.get('threshold'),
                       config.get('signed'),
                       config.get('lut_values_bitwidth', LUT_VALUES_BITWIDTH),
                       config.get('eps', EPS))

else:
    class ActivationLutPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing tensorflow is mandatory '
                         'when using ActivationLutPOTInferableQuantizer. '
                         'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, FOUND_ONNXRUNTIME_EXTENSIONS, ONNX_CUSTOM_OP_DOMAIN
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.common.quant_utils import adjust_range_to_include_zero
from mct_quantizers.logger import Logger
from mct_quantizers.pytorch.onnxruntime_validations import validate_weight_params

if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizers.base_uniform_inferable_quantizer import BaseUniformInferableQuantizer
    from mct_quantizers.pytorch.quantizer_utils import fix_range_to_include_zero, get_working_device, to_torch_tensor
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.base_weight_quantizer_autograd_function import \
        BaseWeightQuantizerAutogradFunction


    def quantize_uniform_weights_torch(input_tensor: torch.Tensor,
                                       num_bits: int,
                                       min_range: np.ndarray,
                                       max_range: np.ndarray,
                                       per_channel: bool,
                                       channel_axis: int=None):
        """
           Quantizes the input tensor uniformly using torch.

           Args:
               input_tensor (torch.Tensor): The input tensor to be quantized.
               num_bits (int): Number of bits to represent the quantized value.
                min_range (np.ndarray): min quantization range for quantizing weights
                max_range (np.ndarray): max quantization range for quantizing weights
                per_channel (bool): Quantize input tensor per-channel or per-tensor.
               channel_axis (int): Axis to quantize the tensor in case of per-channel quantization.

           Returns:
               Uniformly quantized tensor.
        """
        if isinstance(min_range, np.ndarray):
            min_range = torch.tensor(min_range, dtype=torch.float32).to(get_working_device())
        if isinstance(max_range, np.ndarray):
            max_range = torch.tensor(max_range, dtype=torch.float32).to(get_working_device())

        # adjusts the quantization rage so the quantization grid include zero.
        a, b = fix_range_to_include_zero(min_range, max_range, num_bits)

        # Compute the step size of quantized values.
        delta = (b - a) / (2 ** num_bits - 1)

        if per_channel:
            ones = [1] * input_tensor.ndim
            ones[channel_axis] = -1
            new_shape = tuple(ones)
            # Make sure min_values and max_values have the same shape as x along the first axis
            a = torch.reshape(a, new_shape)
            b = torch.reshape(b, new_shape)
            delta = torch.reshape(delta, new_shape)

        # Use torch.where to clip the values in x
        clipped_x = torch.where(input_tensor < a, a, input_tensor)
        quantized = torch.round(torch.where(input_tensor > b, b, clipped_x) / delta) * delta
        return quantized


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    identifier=QuantizerID.INFERABLE)
    class WeightsUniformInferableQuantizer(BaseUniformInferableQuantizer):
        """
        Class for quantizing weights using unsigned uniform quantizer.
        """

        def __init__(self,
                     num_bits: int,
                     min_range: List[float],
                     max_range: List[float],
                     per_channel: bool,
                     channel_axis: int = None
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min quantization range for quantizing weights
                max_range: max quantization range for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.
            """
            super(WeightsUniformInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                   min_range=min_range,
                                                                   max_range=max_range)
            if per_channel:
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert len(min_range) >= 1, f'In per-channel quantization min_range should be of length >= 1 but is {len(min_range)}'
                assert len(max_range) >= 1, f'In per-channel quantization max_range should be of length >= 1 but is {len(max_range)}'
            else:
                assert len(min_range) == 1, f'In per-tensor quantization min_range should be of length 1 but is {len(min_range)}'
                assert len(max_range) == 1, f'In per-tensor quantization max_range should be of length 1 but is {len(max_range)}'

            self.per_channel = per_channel
            self.channel_axis = channel_axis

            self.adjusted_min_range_np = self.min_range.cpu().numpy()
            self.adjusted_max_range_np = self.max_range.cpu().numpy()

            # Compute the step size of quantized values.
            self.scales = (self.max_range - self.min_range) / (2 ** num_bits - 1)
            self.zero_points = -(self.min_range / self.scales).int()  # zp has to be positive, and a <=0, so we multiply by -1

            self.scales = self.scales.to(get_working_device())
            self.zero_points = self.zero_points.to(get_working_device())

        def __call__(self,
                     inputs: torch.Tensor) -> torch.Tensor:
            """
            Weight fake quantizer
            Args:
                inputs: weights to quantize.

            Returns:
                quantized weights
            """
            if self._use_custom_impl and torch.jit.is_tracing():
                return WeightsUniformF.apply(inputs,
                                             self.num_bits,
                                             self.adjusted_min_range_np,
                                             self.adjusted_max_range_np,
                                             self.per_channel,
                                             self.channel_axis)

            inputs.requires_grad = False
            if self.per_channel:
                return torch.fake_quantize_per_channel_affine(inputs,
                                                              self.scales.flatten(),
                                                              self.zero_points.flatten(),
                                                              axis=self.channel_axis,
                                                              quant_min=self.min_quantized_domain,
                                                              quant_max=self.max_quantized_domain)
            return torch.fake_quantize_per_tensor_affine(inputs,
                                                         self.scales,
                                                         self.zero_points,
                                                         quant_min=self.min_quantized_domain,
                                                         quant_max=self.max_quantized_domain)


    class WeightsUniformF(BaseWeightQuantizerAutogradFunction):
        """
        Custom autograd function for uniform weights quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, num_bits, min_range, max_range, per_channel, channel_axis):
            """
             Forward computation function. This method performs the forward computation using
             the given quantize_sym_weights_torch function.

             Args:
                 ctx: An object that can be used to stash information for backward function.
                 input_tensor: The input tensor to be quantized.
                 num_bits: The number of bits to represent the quantized tensor.
                 min_range: min quantization range for quantizing weights
                 max_range: max quantization range for quantizing weights
                 per_channel: whether to use per-channel quantization
                 channel_axis: Axis of input to apply per-channel quantization on.

             Returns:
                 The quantized tensor.
             """
            return quantize_uniform_weights_torch(input_tensor, num_bits, min_range, max_range, per_channel,
                                                  channel_axis)

        @staticmethod
        def symbolic(g, input_tensor, num_bits, min_range, max_range, per_channel, channel_axis):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                num_bits: The number of bits to represent the quantized value.
                min_range: min quantization range for quantizing weights
                max_range: max quantization range for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            # When None is passed as channel_axis, the op has no attribute of channel_axis,
            # which creates conflict with the onnxruntime function. For this reason, if we quantize
            # per-tensor and channel_axis is None, we set it to 0.
            if not per_channel and channel_axis is None:
                channel_axis = 0

            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsUniformQuantizer", input_tensor,
                        g.op('Constant', value_t=torch.tensor(min_range, dtype=torch.float32)),
                        g.op('Constant', value_t=torch.tensor(max_range, dtype=torch.float32)),
                        num_bits_i=num_bits,
                        per_channel_i=int(per_channel),
                        channel_axis_i=channel_axis,
                        signed_i=WeightsUniformF.is_signed(),
                        **WeightsUniformF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())


else:
    class WeightsUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing torch is mandatory '
                         'when using WeightsUniformInferableQuantizer. '
                         'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef

    def quantize_uniform_weights_numpy(input_tensor: np.ndarray,
                                       num_bits: int,
                                       min_range: np.ndarray,
                                       max_range: np.ndarray,
                                       per_channel: bool,
                                       channel_axis: int=None):
        """
           Quantizes the input tensor uniformly using numpy.

           Args:
               input_tensor (np.ndarray): The input tensor to be quantized.
               num_bits (int): Number of bits to represent the quantized value.
                min_range (np.ndarray): min quantization range for quantizing weights
                max_range (np.ndarray): max quantization range for quantizing weights
                per_channel (bool): Quantize input tensor per-channel or per-tensor.
               channel_axis (int): Axis to quantize the tensor in case of per-channel quantization.

           Returns:
               Uniformly quantized tensor.
        """

        validate_weight_params(input_tensor=input_tensor,
                               per_channel=per_channel,
                               min_range=min_range,
                               max_range=max_range,
                               channel_axis=channel_axis)

        # adjusts the quantization rage so the quantization grid include zero.
        a, b = adjust_range_to_include_zero(min_range, max_range, num_bits)

        # Compute the step size of quantized values.
        delta = (b - a) / (2 ** num_bits - 1)
        if per_channel:
            ones = np.ones(input_tensor.ndim)
            ones[channel_axis] = -1
            new_shape = tuple([int(x) for x in ones])
            # Make sure min_values and max_values have the same shape as x along the first axis
            a = np.reshape(a, new_shape)
            b = np.reshape(b, new_shape)
            delta = np.reshape(delta, new_shape)

        # Use torch.where to clip the values in x
        clipped_x = np.where(input_tensor < a, a, input_tensor)
        quantized = np.round(np.where(input_tensor > b, b, clipped_x) / delta) * delta
        return quantized

    # Add onnx op function to use during onnxruntime WeightsUniformQuantizer op inference
    # Using this decorator the op WeightsUniformQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsUniformQuantizer",
             inputs=[PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float
                     ],
             outputs=[PyCustomOpDef.dt_float],
             attrs={
                 "num_bits": PyCustomOpDef.dt_int64,
                 "per_channel": PyCustomOpDef.dt_int64,
                 "channel_axis": PyCustomOpDef.dt_int64,
             }
             )
    def weight_uniform_ort(x, min_range, max_range, **kwargs):
        return quantize_uniform_weights_numpy(x,
                                              kwargs["num_bits"],
                                              min_range,
                                              max_range,
                                              kwargs["per_channel"],
                                              kwargs["channel_axis"]
                                              )



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, FOUND_ONNXRUNTIME_EXTENSIONS, ONNX_CUSTOM_OP_DOMAIN
from mct_quantizers.common.quant_info import QuantizationMethod


if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.weights_symmetric_inferable_quantizer import \
        WeightsSymmetricInferableQuantizer, quantize_sym_weights_torch
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.base_weight_quantizer_autograd_function import \
        BaseWeightQuantizerAutogradFunction


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.POWER_OF_TWO],
                    identifier=QuantizerID.INFERABLE)
    class WeightsPOTInferableQuantizer(WeightsSymmetricInferableQuantizer):
        """
        Class for quantizing weights using a power-of-two quantizer.
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     ):

            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.
            """
            # target of Weights quantization
            super(WeightsPOTInferableQuantizer, self).__init__(num_bits=num_bits,
                                                               threshold=threshold,
                                                               per_channel=per_channel,
                                                               channel_axis=channel_axis)
            self.num_bits = num_bits
            self.threshold = threshold
            self.per_channel = per_channel
            self.channel_axis = channel_axis

            is_threshold_pot = np.all(
                np.round(np.log2(self.threshold_np.flatten())) == np.log2(self.threshold_np.flatten()))
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {threshold}'

        def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """

            if self._use_custom_impl and torch.jit.is_tracing():
                return WeightsPOTF.apply(inputs,
                                         self.num_bits,
                                         self.threshold_np,
                                         self.per_channel,
                                         self.channel_axis)

            return super(WeightsPOTInferableQuantizer, self).__call__(inputs)


    class WeightsPOTF(BaseWeightQuantizerAutogradFunction):
        """
        Custom autograd function for POT weights quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, num_bits, threshold, per_channel, channel_axis):
            """
             Forward computation function. This method performs the forward computation using
             the given quantize_sym_weights_torch function.

             Args:
                 ctx: An object that can be used to stash information for backward function.
                 input_tensor: The input tensor to be quantized.
                 num_bits: The number of bits to represent the quantized tensor.
                 threshold: The quantization threshold.
                 per_channel: whether to use per-channel quantization
                 channel_axis: Axis of input to apply per-channel quantization on.

             Returns:
                 The quantized tensor.
             """
            return quantize_sym_weights_torch(input_tensor, num_bits, threshold, per_channel, channel_axis)

        @staticmethod
        def symbolic(g, input_tensor, num_bits, threshold, per_channel, channel_axis):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                num_bits: The number of bits to represent the quantized value.
                threshold: The quantization threshold.
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            # When None is passed as channel_axis, the op has no attribute of channel_axis,
            # which creates conflict with the onnxruntime function. For this reason, if we quantize
            # per-tensor and channel_axis is None, we set it to 0.
            if not per_channel and channel_axis is None:
                channel_axis = 0

            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsPOTQuantizer", input_tensor,
                        g.op('Constant', value_t=torch.tensor(threshold, dtype=torch.float32)),
                        num_bits_i=num_bits,
                        per_channel_i=int(per_channel),
                        channel_axis_i=channel_axis,
                        signed_i=int(WeightsPOTF.is_signed()),
                        **WeightsPOTF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())


else:
    class WeightsPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using WeightsPOTInferableQuantizer. '
                            'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.weights_symmetric_inferable_quantizer import \
        quantize_sym_weights_numpy


    # Add onnx op function to use during onnxruntime WeightsPOTQuantizer op inference
    # Using this decorator the op WeightsPOTQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsPOTQuantizer",
             inputs=[PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={
                 "num_bits": PyCustomOpDef.dt_int64,
                 "per_channel": PyCustomOpDef.dt_int64,
                 "channel_axis": PyCustomOpDef.dt_int64,
             }
             )
    def weight_pot_ort(input_tensor: np.ndarray, threshold: np.ndarray, **kwargs):
        return quantize_sym_weights_numpy(input_tensor,
                                          kwargs["num_bits"],
                                          threshold,
                                          kwargs["per_channel"],
                                          kwargs["channel_axis"])



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List
import numpy as np

from mct_quantizers.common.base_inferable_quantizer import QuantizationTarget, mark_quantizer, QuantizerID
from mct_quantizers.common.constants import FOUND_TF
from mct_quantizers.common.quant_info import QuantizationMethod

if FOUND_TF:
    from mct_quantizers.keras.quantizers.weights_inferable_quantizers.weights_symmetric_inferable_quantizer import \
        WeightsSymmetricInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.POWER_OF_TWO],
                    identifier=QuantizerID.INFERABLE)
    class WeightsPOTInferableQuantizer(WeightsSymmetricInferableQuantizer):
        """
        Class for quantizing weights using a power-of-two quantizer.
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                per_channel: whether to use per-channel quantization
                channel_axis: axis along which to apply per-channel quantization
                input_rank: number of dimensions of input tensor the quantizer quantizes
            """
            # Call the superclass constructor with the given parameters, along with the target of Weights quantization
            super(WeightsPOTInferableQuantizer, self).__init__(num_bits=num_bits,
                                                               threshold=threshold,
                                                               per_channel=per_channel,
                                                               channel_axis=channel_axis,
                                                               input_rank=input_rank)

            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in self._np_threshold.flatten()])
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {self.threshold}'

        @classmethod
        def from_config(cls, config):
            """
            Return an object with config
            Args:
                config(dict): dictionary of object configuration
            Returns: An object created with config
            """
            return cls(config.get('num_bits'),
                       config.get('threshold'),
                       config.get('per_channel'),
                       config.get('channel_axis', None),
                       config.get('input_rank', None))


else:
    class WeightsPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing tensorflow is mandatory '
                            'when using WeightsPOTInferableQuantizer. '
                            'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict

import torch
from mct_quantizers import __version__ as mctq_version
from mct_quantizers.common.constants import MCTQ_VERSION


class BaseQuantizerAutogradFunction(torch.autograd.Function):
    """
    Custom autograd function for quantizer.
    It provides a way to define a custom forward and symbolic operation
    and currently does not implement a backward operation.
    """

    @staticmethod
    def forward(ctx, input_tensor, **kwargs):
        """
         Forward computation function. This method performs the forward computation using
         the given quantize_sym_weights_torch function.
         """
        raise NotImplemented

    @staticmethod
    def symbolic(g, input_tensor, **kwargs):
        raise NotImplemented

    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        """
        Backward computation function. Raises a NotImplementedError
        since backward is not needed for this op.

        Args:
            ctx (Any): A context object from the forward pass.
            grad_outputs (Any): Gradients w.r.t. the output tensor.
        """
        raise NotImplementedError()

    @staticmethod
    def _get_metadata_attributes() -> Dict[str,Any]:
        """

        Returns: Metadata dictionary for all quantizers onnx symbolic ops.

        """
        return {f"{MCTQ_VERSION}_s": mctq_version}


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, LUT_VALUES_BITWIDTH, EPS, ONNX_CUSTOM_OP_DOMAIN, \
    FOUND_ONNXRUNTIME_EXTENSIONS
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.common.quant_utils import lut_quantizer_np

if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizer_utils import to_torch_tensor, get_working_device, lut_quantizer
    from mct_quantizers.pytorch.quantizers.base_lut_symmetric_inferable_quantizer import \
        BaseLUTSymmetricInferableQuantizer
    from mct_quantizers.pytorch.quantizers.weights_inferable_quantizers.base_weight_quantizer_autograd_function import \
        BaseWeightQuantizerAutogradFunction


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.LUT_SYM_QUANTIZER],
                    identifier=QuantizerID.INFERABLE)
    class WeightsLUTSymmetricInferableQuantizer(BaseLUTSymmetricInferableQuantizer):
        """
        Class for quantizing weights using a lut symmetric quantizer
        """

        def __init__(self,
                     num_bits: int,
                     lut_values: List[float],
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None,
                     lut_values_bitwidth: int = LUT_VALUES_BITWIDTH,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                lut_values: the values in the look-up table to assign the weights to
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on
                lut_values_bitwidth: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """

            super(WeightsLUTSymmetricInferableQuantizer, self).__init__(threshold=threshold,
                                                                        num_bits=num_bits,
                                                                        lut_values=lut_values,
                                                                        signed=True,
                                                                        lut_values_bitwidth=lut_values_bitwidth,
                                                                        eps=eps)

            self.per_channel = per_channel
            self.channel_axis = channel_axis
            self.input_rank = input_rank

            if per_channel:
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert input_rank is not None, f'input_rank is missing in per channel quantization'
                assert len(
                    threshold) >= 1, f'In per-channel quantization threshold should be of length >= 1 but is ' \
                                     f'{len(threshold)}'
            else:
                assert len(
                    threshold) == 1, f'In per-tensor quantization threshold should be of length 1 but is ' \
                                     f'{len(threshold)}'

            self._threshold_torch = to_torch_tensor(self._threshold_np).to(get_working_device())
            self._lut_values_torch = to_torch_tensor(self._lut_values_np).to(get_working_device())

        def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            if self._use_custom_impl and torch.jit.is_tracing():
                return WeightsLUTSymmetricF.apply(inputs,
                                                  self.num_bits,
                                                  self._lut_values_np,
                                                  self._threshold_np,
                                                  self.lut_values_bitwidth,
                                                  self.eps,
                                                  self.per_channel,
                                                  self.channel_axis,
                                                  self.input_rank
                                                  )

            inputs.requires_grad = False
            return lut_quantizer(inputs,
                                 lut_values=self._lut_values_torch,
                                 signed=True,
                                 threshold=self._threshold_torch,
                                 lut_values_bitwidth=self.lut_values_bitwidth,
                                 eps=self.eps,
                                 per_channel=self.per_channel,
                                 channel_axis=self.channel_axis,
                                 input_rank=self.input_rank
                                 )


    class WeightsLUTSymmetricF(BaseWeightQuantizerAutogradFunction):
        """
        Custom autograd function for symmetric weights quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, num_bits, lut_values, threshold, lut_values_bitwidth, eps,
                    per_channel,
                    channel_axis,
                    input_rank
                    ):
            # Apply pass np arrays to support the symbolic op, so it needs to be converted to
            # torch since here we quantize a torch tensor
            threshold_torch = to_torch_tensor(threshold).to(get_working_device())
            lut_values_torch = to_torch_tensor(lut_values).to(get_working_device())

            return lut_quantizer(input_tensor,
                                 lut_values=lut_values_torch,
                                 signed=True,
                                 threshold=threshold_torch,
                                 lut_values_bitwidth=lut_values_bitwidth,
                                 eps=eps,
                                 per_channel=per_channel,
                                 channel_axis=channel_axis,
                                 input_rank=input_rank)

        @staticmethod
        def symbolic(g,
                     input_tensor,
                     num_bits,
                     lut_values,
                     threshold,
                     lut_values_bitwidth,
                     eps,
                     per_channel,
                     channel_axis,
                     input_rank):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                num_bits: The number of bits to represent the quantized value.
                threshold: The quantization threshold.
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            # When None is passed as channel_axis, the op has no attribute of channel_axis,
            # which creates conflict with the onnxruntime function. For this reason, if we quantize
            # per-tensor and channel_axis is None, we set it to 0.
            # per-tensor and input_rank is None, we set it to 4.
            if not per_channel and channel_axis is None:
                channel_axis = 0
            if not per_channel and input_rank is None:
                input_rank = 4

            return g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsLUTSymmetricQuantizer", input_tensor,
                        g.op('Constant', value_t=torch.tensor(lut_values, dtype=torch.float32)),
                        g.op('Constant', value_t=torch.tensor(threshold, dtype=torch.float32)),
                        num_bits_i=num_bits,
                        per_channel_i=int(per_channel),
                        channel_axis_i=channel_axis,
                        input_rank_i=input_rank,
                        lut_values_bitwidth_i=lut_values_bitwidth,
                        eps_f=eps,
                        signed_i=int(WeightsLUTSymmetricF.is_signed()),
                        **WeightsLUTSymmetricF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())


else:
    class WeightsLUTSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using WeightsLUTSymmetricInferableQuantizer. '
                            'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from onnxruntime_extensions import onnx_op, PyCustomOpDef
    def quantize_lut_sym_weights_numpy(input_tensor: np.ndarray,
                                       lut_values,
                                       threshold,
                                       lut_values_bitwidth,
                                       eps,
                                       per_channel,
                                       channel_axis=None,
                                       input_rank=None
                                       ):
        quantized_tensor = lut_quantizer_np(tensor_data=input_tensor,
                                            lut_values=lut_values,
                                            signed=True,
                                            threshold=threshold,
                                            lut_values_bitwidth=lut_values_bitwidth,
                                            eps=eps,
                                            per_channel=per_channel,
                                            channel_axis=channel_axis,
                                            input_rank=input_rank)
        return quantized_tensor


    # Add onnx op function to use during onnxruntime WeightsLUTSymmetricQuantizer op inference
    # Using this decorator the op WeightsLUTSymmetricQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::WeightsLUTSymmetricQuantizer",
             inputs=[PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float,
                     PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={
                 "lut_values_bitwidth": PyCustomOpDef.dt_int64,
                 "eps": PyCustomOpDef.dt_float,
                 "per_channel": PyCustomOpDef.dt_int64,
                 "channel_axis": PyCustomOpDef.dt_int64,
                 "input_rank": PyCustomOpDef.dt_int64
             })
    def weight_lut_sym_ort(input_tensor: np.ndarray,
                           lut_values: np.ndarray,
                           threshold: np.ndarray,
                           **kwargs):
        return quantize_lut_sym_weights_numpy(input_tensor,
                                              lut_values,
                                              threshold,
                                              kwargs["lut_values_bitwidth"],
                                              kwargs["eps"],
                                              kwargs["per_channel"],
                                              kwargs["channel_axis"],
                                              kwargs["input_rank"])



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import warnings
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH
from mct_quantizers.common.quant_info import QuantizationMethod

if FOUND_TORCH:
    from mct_quantizers.pytorch.quantizers.base_pytorch_inferable_quantizer import BasePyTorchInferableQuantizer

    @mark_quantizer(quantization_target=None,
                    quantization_method=[QuantizationMethod.LUT_SYM_QUANTIZER],
                    identifier=QuantizerID.INFERABLE)
    class BaseLUTSymmetricInferableQuantizer(BasePyTorchInferableQuantizer):

        def __init__(self,
                     num_bits: int,
                     lut_values: List[float],
                     threshold: List[float],
                     signed: bool,
                     lut_values_bitwidth: int,
                     eps: float):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                lut_values: the values in the look-up table to assign the weights to
                threshold: threshold for quantizing values
                signed: whether or not to use signed quantization
                lut_values_bitwidth: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """

            super(BaseLUTSymmetricInferableQuantizer, self).__init__()

            assert isinstance(threshold,
                              list), f'Threshold is expected to be a list, but is of type {type(threshold)}'
            assert isinstance(lut_values,
                              list), f'lut_values is expected to be a list, but is of type {type(lut_values)}'


            self._threshold_np = np.asarray(threshold)
            self._lut_values_np = np.asarray(lut_values)

            assert len(np.unique(self._lut_values_np)) <= 2 ** num_bits, \
                f'Expected num of lut values to be less or equal than {2 ** num_bits} ' \
                f'but got {len(self._lut_values_np)}'

            assert not np.any(self._lut_values_np - self._lut_values_np.astype(int)), f'Expected lut values to be integers'

            if signed:
                assert np.all((-1 * (2 ** (lut_values_bitwidth - int(signed))) <= self._lut_values_np) &
                              (self._lut_values_np <= (2 ** (lut_values_bitwidth - int(signed)) - 1))), \
                    f'Expected lut values in the quantization range'
            else:
                assert np.all(self._lut_values_np <= (2 ** lut_values_bitwidth)), f'Expected lut values in the ' \
                                                                            f'quantization range'

            # If unsigned activation quantization, all lut_values must be positive
            if not signed:
                assert np.all(self._lut_values_np >= 0), f'Expected unsigned lut values in unsigned activation ' \
                                                          f'quantization'

            # num_bits must be less than lut_values_bitwidth
            assert num_bits <= lut_values_bitwidth, f'Look-Up-Table bit configuration has {num_bits} bits. It must be ' \
                                                  f'less then {lut_values_bitwidth}'
            if num_bits == lut_values_bitwidth:
                warnings.warn("Num of bits equal to multiplier n bits, Please be aware LUT quantizier may be "
                              "inefficient in that case, consider using SymmetricInferableQuantizer instead")

            self.threshold = threshold
            self.lut_values = lut_values

            self.signed = signed
            self.num_bits = num_bits
            self.lut_values_bitwidth = lut_values_bitwidth
            self.eps = eps

else:
    class BaseLUTSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory when using BaseLUTSymmetricInferableQuantizer. Could not '
                            'find torch package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np


def validate_weight_params(input_tensor: np.ndarray,
                           min_range: np.ndarray,
                           max_range: np.ndarray,
                           per_channel: int,
                           channel_axis: int
                           ):
    """
    Validate onnxruntime weight quantization function params.
    """

    assert isinstance(input_tensor,
                      np.ndarray), f"Input tensor expected to be a numpy array but is {type(input_tensor)}"

    assert isinstance(min_range,
                      np.ndarray), f"min_range expected to be a numpy array but is {type(min_range)}"
    assert isinstance(max_range,
                      np.ndarray), f"max_range expected to be a numpy array but is {type(max_range)}"

    assert min_range.ndim == 1, f"min_range ndim expected to be 1 but is {min_range.ndim}"
    assert max_range.ndim == 1, f"max_range ndim expected to be 1 but is {max_range.ndim}"

    assert min_range.shape[0] == max_range.shape[
        0], f'Expected min and max ranges to have same shapes but min_range shape: {min_range.shape} and max_range ' \
            f'shape: {max_range.shape}'

    assert np.all(
        min_range < max_range), f"max_range should be greater than min_range but max: {max_range}, min: {min_range}"

    if per_channel:
        assert channel_axis is not None, f'In per channel quantization channel_axis must be provided'
        assert input_tensor.shape[channel_axis] == min_range.shape[
            0], f"Mismatch between channels to quantize {input_tensor.shape[channel_axis]} to number of quantization " \
                f"ranges: " \
                f"{min_range.shape[0]}"
    else:
        assert min_range.shape[0] == 1, f'In per tensor quantization only one quantization range should be provided'


def validate_activation_params(input_tensor: np.ndarray,
                               min_range: float,
                               max_range: float):
    """
    Validate onnxruntime activation quantization function params.
    """

    assert isinstance(input_tensor,
                      np.ndarray), f"Input tensor expected to be a numpy array but is {type(input_tensor)}"

    assert isinstance(min_range,
                      float), f"min_range expected to be float but is {type(min_range)}"
    assert isinstance(max_range,
                      float), f"max_range expected to be float but is {type(max_range)}"
    assert min_range < max_range, f"max_range should be greate than min_range but max: {max_range}, min: {min_range}"



# Copyright 2024 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Any
import sys

from mct_quantizers import __version__ as mctq_version
from mct_quantizers.logger import Logger
from mct_quantizers.common.constants import MCTQ_VERSION, PYTHON_VERSION


def verify_and_init_metadata(metadata: Dict = None):
    """
    Init the metadata dictionary and verify its compliance

    Args:
        metadata (Dict): metadata dictionary. Should contain only string keys and string\interger\float values or
        list/dictionaries with these allowed data types.

    Returns:
        The metadata dictionary with added default values.

    """

    def _validate_metadata_value(value: Any) -> bool:
        """
        Validates that the value is either an int, float, str, list, or dict.
        For lists, all elements must be int, float, or str.
        For dicts, all keys must be str and all values must be int, float, str, list, or dict.

        Args:
            value (Any): The value to be validated.

        Returns:
            bool: True if the value is valid, False otherwise.
        """
        if isinstance(value, (int, float, str)):
            return True
        elif isinstance(value, list):
            return all(_validate_metadata_value(item) for item in value)
        elif isinstance(value, dict):
            return all(isinstance(k, str) and _validate_metadata_value(v) for k, v in value.items())
        return False

    if not isinstance(metadata, dict):
        Logger.error(f'metadata should be a dictionary, but got type {type(metadata)}.')
    if not all(isinstance(k, str) for k in metadata.keys()):
        Logger.error('metadata dictionary should only have string keys.')
    if not all(_validate_metadata_value(v) for v in metadata.values()):
        Logger.warning('metadata dictionary values should be strings, integers, floats, lists, '
                       'or dictionaries with appropriate inner values. Other types may cause issues '
                       'with saving/loading the metadata.')

    if metadata is None:
        metadata = {}
    if PYTHON_VERSION not in metadata:
        metadata[PYTHON_VERSION] = sys.version
    if MCTQ_VERSION not in metadata:
        metadata[MCTQ_VERSION] = mctq_version

    return metadata



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TF
from mct_quantizers.common.quant_info import QuantizationMethod


if FOUND_TF:
    import tensorflow as tf
    from mct_quantizers.keras.quantizers.activation_inferable_quantizers.activation_uniform_inferable_quantizer \
        import ActivationUniformInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.SYMMETRIC],
                    identifier=QuantizerID.INFERABLE)
    class ActivationSymmetricInferableQuantizer(ActivationUniformInferableQuantizer):

        """
        Class for quantizing activations using a symmetric quantizer
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     signed: bool):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                signed: whether or not to use signed quantization
            """
            assert isinstance(threshold, list), f'Expected threshold to be of type list but is {type(threshold)}'
            # In activation per-channel quantization is not supported thus we expect a single min/max value.
            assert len(threshold) == 1, f'In per-tensor quantization threshold should be of length 1 but is {len(threshold)}'
            assert all([isinstance(x, (float, np.float32, tf.float32)) for x in
                        threshold]), f'Expected threshold list to contain float or np.float values but found ' \
                                     f'{[type(x) for x in threshold]}'

            self.threshold = threshold
            self.signed = signed

            delta = self.threshold[0] / (2 ** (num_bits - int(self.signed)))
            # In activation quantization is per-tensor only - thus we pass the threshold as a list with a len of 1
            min_range = [-threshold[0]] if self.signed else [0.0]
            max_range = [self.threshold[0] - delta]

            super(ActivationSymmetricInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                        min_range=min_range,
                                                                        max_range=max_range)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'signed', 'threshold'
            """
            return {'num_bits': self.num_bits,
                    'signed': self.signed,
                    'threshold': self.threshold}

        @classmethod
        def from_config(cls, config):
            """
            Return an object with config
            Args:
                config(dict): dictionary of object configuration
            Returns: An object created with config
            """
            return cls(config.get('num_bits'),
                       config.get('threshold'),
                       config.get('signed'))
else:
    class ActivationSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing tensorflow is mandatory '
                            'when using ActivationSymmetricInferableQuantizer. '
                            'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH, FOUND_ONNXRUNTIME_EXTENSIONS, ONNX_CUSTOM_OP_DOMAIN
from mct_quantizers.common.quant_info import QuantizationMethod

if FOUND_TORCH:
    import torch
    from mct_quantizers.pytorch.quantizers.activation_inferable_quantizers.activation_symmetric_inferable_quantizer import ActivationSymmetricInferableQuantizer, quantize_sym_activations_torch
    from mct_quantizers.pytorch.quantizers.activation_inferable_quantizers.base_activation_quantizer_autograd_function import BaseActivationQuantizerAutogradFunction

    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.POWER_OF_TWO],
                    identifier=QuantizerID.INFERABLE)
    class ActivationPOTInferableQuantizer(ActivationSymmetricInferableQuantizer):
        """
        Class for quantizing activations using power-of-two quantizer
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     signed: bool):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                signed: whether to use signed quantization or not
            """
            # target of Activation quantization
            super(ActivationPOTInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                  signed=signed,
                                                                  threshold=threshold)

            is_threshold_pot = np.all(
                np.round(np.log2(self.threshold_np.flatten())) == np.log2(self.threshold_np.flatten()))
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {threshold}'

        def __call__(self, inputs):
            """
            Quantize an input tensor. If custom implementation is enabled, quantizer's autograd
            function class is applied.

            Args:
                inputs: Activation tensor to quantize.

            Returns:
                Quantized tensor.
            """
            if self._use_custom_impl and torch.jit.is_tracing():
                return ActivationPOTF.apply(inputs,
                                            self.threshold_np,
                                            self.signed,
                                            self.num_bits)
            return super(ActivationPOTInferableQuantizer, self).__call__(inputs)


    class ActivationPOTF(BaseActivationQuantizerAutogradFunction):
        """
        Custom autograd function for POT activations quantizer.
        It provides a way to define a custom forward and symbolic operation
        and currently does not implement a backward operation.
        """

        @staticmethod
        def forward(ctx, input_tensor, threshold, signed, num_bits):
            """
             Forward computation function. This method performs the forward computation using
             the given quantize_sym_activations_torch function.

             Args:
                 ctx: An object that can be used to stash information for backward function.
                 input_tensor: The input tensor to be quantized.
                 threshold: The quantization threshold.
                 signed: A flag that indicates if the quantization is signed or unsigned.
                 num_bits: The number of bits to represent the quantized tensor.

             Returns:
                 The quantized tensor.
             """
            return quantize_sym_activations_torch(input_tensor, threshold, signed, num_bits)

        @staticmethod
        def symbolic(g, input_tensor, threshold, signed, num_bits):
            """
            Symbolic method that defines the custom operation for ONNX export.

            Args:
                g: A graph object that represents the ONNX computation graph.
                input_tensor: The input tensor to be quantized.
                threshold: The quantization threshold.
                signed: A flag that indicates if the quantization is signed or unsigned.
                num_bits: The number of bits to represent the quantized value.

            Returns:
                The node in the ONNX graph representing the output of this operation.
            """
            added_op = g.op(f"{ONNX_CUSTOM_OP_DOMAIN}::ActivationPOTQuantizer",
                        input_tensor,
                        threshold_f=threshold,
                        signed_i=int(signed),
                        num_bits_i=num_bits,
                            **ActivationPOTF._get_metadata_attributes()
                        ).setType(
                input_tensor.type())
            return added_op


else:
    class ActivationPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using ActivationPOTInferableQuantizer. '
                            'Could not find torch package.')

if FOUND_ONNXRUNTIME_EXTENSIONS:
    from mct_quantizers.pytorch.quantizers.activation_inferable_quantizers.activation_symmetric_inferable_quantizer import quantize_sym_activations_numpy
    from onnxruntime_extensions import onnx_op, PyCustomOpDef

    # Add onnx op function to use during onnxruntime ActivationPOTQuantizer op inference.
    # Using this decorator the op ActivationPOTQuantizer is defined using its inputs, outputs and attributes.
    @onnx_op(op_type=f"{ONNX_CUSTOM_OP_DOMAIN}::ActivationPOTQuantizer",
             inputs=[PyCustomOpDef.dt_float],
             outputs=[PyCustomOpDef.dt_float],
             attrs={"threshold": PyCustomOpDef.dt_float,
                    "signed": PyCustomOpDef.dt_int64,
                    "num_bits": PyCustomOpDef.dt_int64
                    })
    def activation_pot_ort(input_tensor,
                           **kwargs):
        return quantize_sym_activations_numpy(input_tensor,
                                              kwargs["threshold"],
                                              kwargs["signed"],
                                              kwargs["num_bits"]
                                              )



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from mct_quantizers.common.base_inferable_quantizer import mark_quantizer, QuantizerID
from mct_quantizers.common.constants import FOUND_TORCH
from mct_quantizers.common.quant_info import QuantizationMethod
from mct_quantizers.pytorch.quantizer_utils import get_working_device, to_torch_tensor, fix_range_to_include_zero

if FOUND_TORCH:
    from mct_quantizers.pytorch.quantizers.base_pytorch_inferable_quantizer import BasePyTorchInferableQuantizer


    @mark_quantizer(quantization_target=None,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    identifier=QuantizerID.INFERABLE)
    class BaseUniformInferableQuantizer(BasePyTorchInferableQuantizer):

        def __init__(self,
                     num_bits: int,
                     min_range: List[float],
                     max_range: List[float]):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min quantization range for quantizing
                max_range: max quantization range for quantizing
            """

            super(BaseUniformInferableQuantizer, self).__init__()

            assert isinstance(min_range, list), f'min_range is expected to be a list, but is of type {type(min_range)}'
            assert isinstance(max_range, list), f'max_range is expected to be a list, but is of type {type(max_range)}'

            for _min, _max in zip(min_range, max_range):
                assert _min<_max, f"Max range must be greater than min value but min is {_min} and max is {_max}"

            # Align mix/max numpy arrays so they are torch Tensors on the working device
            min_range = to_torch_tensor(np.asarray(min_range)).to(get_working_device())
            max_range = to_torch_tensor(np.asarray(max_range)).to(get_working_device())

            min_range, max_range = fix_range_to_include_zero(min_range,
                                                             max_range,
                                                             num_bits)
            self.min_range = min_range
            self.max_range = max_range

            self.num_bits = num_bits
            self.min_quantized_domain = 0
            self.max_quantized_domain = 2 ** num_bits - 1


else:
    class BaseUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using BaseUniformInferableQuantizer. '
                            'Could not find torch package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from mct_quantizers.common.base_inferable_quantizer import BaseInferableQuantizer
from mct_quantizers.common.constants import ACTIVATION_HOLDER_QUANTIZER, FOUND_TF, TRAINING, STEPS, MCTQ_VERSION
from mct_quantizers.common.get_all_subclasses import get_all_subclasses
from mct_quantizers.logger import Logger
from mct_quantizers import __version__ as mctq_version

if FOUND_TF:
    import tensorflow as tf
    from tensorflow.python.util import tf_inspect
    from tensorflow.python.keras.utils.control_flow_util import smart_cond
    from mct_quantizers.keras.quantizers import BaseKerasInferableQuantizer

    keras = tf.keras


    def _make_quantizer_fn(quantizer, x, training):
        """Use currying to return True/False specialized fns to the cond."""

        def quantizer_fn():
            return quantizer(x, training)

        return quantizer_fn


    class KerasActivationQuantizationHolder(keras.layers.Layer):
        """
        Keras layer to hold an activation quantizer and quantize during inference.
        """
        def __init__(self,
                     activation_holder_quantizer: BaseInferableQuantizer,
                     **kwargs):
            """

            Args:
                activation_holder_quantizer: Quantizer to use during inference.
                **kwargs: Key-word arguments for the base layer
            """

            super(KerasActivationQuantizationHolder, self).__init__(**kwargs)
            self.activation_holder_quantizer = activation_holder_quantizer

            self._mctq_version = mctq_version

        def get_config(self):
            """
            config(dict): dictionary  of ActivationQuantizationHolder Configuration

            Returns: Configuration of ActivationQuantizationHolder.

            """
            base_config = super(KerasActivationQuantizationHolder, self).get_config()
            config = {
                ACTIVATION_HOLDER_QUANTIZER: keras.utils.serialize_keras_object(self.activation_holder_quantizer)}

            return_config = dict(list(base_config.items()) + list(config.items()))
            return_config[MCTQ_VERSION] = self._mctq_version

            return return_config

        @classmethod
        def from_config(cls, config):
            """

            Args:
                config(dict): dictionary  of  ActivationQuantizationHolder Configuration

            Returns: A ActivationQuantizationHolder object

            """
            qi_inferable_custom_objects = {subclass.__name__: subclass for subclass in
                                           get_all_subclasses(BaseKerasInferableQuantizer)}
            config = config.copy()
            activation_holder_quantizer = keras.utils.deserialize_keras_object(config.pop(ACTIVATION_HOLDER_QUANTIZER),
                                                                               module_objects=globals(),
                                                                               custom_objects=qi_inferable_custom_objects)

            v = config.pop(MCTQ_VERSION, None)

            obj = cls(activation_holder_quantizer=activation_holder_quantizer, **config)
            obj._mctq_version = mctq_version if v is None else v

            return obj

        @property
        def mctq_version(self):
            return self._mctq_version

        def build(self, input_shape):
            """
            ActivationQuantizationHolder build function.
            Args:
                input_shape: the layer input shape

            Returns: None

            """
            super(KerasActivationQuantizationHolder, self).build(input_shape)

            self.optimizer_step = self.add_weight(
                STEPS,
                initializer=tf.keras.initializers.Constant(-1),
                dtype=tf.dtypes.int32,
                trainable=False)

            self.activation_holder_quantizer.initialize_quantization(None,
                                                                     self.name + '/out_',
                                                                     self)

        def call(self,
                 inputs: tf.Tensor,
                 training=None) -> tf.Tensor:
            """
            Quantizes the input tensor using the activation quantizer the ActivationQuantizationHolder holds.

            Args:
                inputs: Input tensors to quantize use the activation quantizer the object holds
                training: a boolean stating if layer is in training mode.

            Returns: Output of the activation quantizer (quantized input tensor).

            """
            if training is None:
                training = tf.keras.backend.learning_phase()

            activation_quantizer_args_spec = tf_inspect.getfullargspec(self.activation_holder_quantizer.__call__).args
            if TRAINING in activation_quantizer_args_spec:
                return smart_cond(
                    training,
                    _make_quantizer_fn(self.activation_holder_quantizer, inputs, True),
                    _make_quantizer_fn(self.activation_holder_quantizer, inputs, False))

            return self.activation_holder_quantizer(inputs)

        def convert_to_inferable_quantizers(self):
            """
            Convert layer's quantizer to inferable quantizer.

            Returns:
                None
            """
            if hasattr(self.activation_holder_quantizer, 'convert2inferable') and callable(
                    self.activation_holder_quantizer.convert2inferable):  # pragma: no cover
                self.activation_holder_quantizer = self.activation_holder_quantizer.convert2inferable()
                return self.from_config(self.get_config()) # return new layer with no weights. It assumes holder of inferable quantizers have no weights


else:
    class KerasActivationQuantizationHolder:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing tensorflow is mandatory '
                         'when using ActivationQuantizationHolder. '
                         'Could not find Tensorflow package.')



# Copyright 2024 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict

from mct_quantizers.common.constants import FOUND_TORCH, FRAMEWORK_VERSION, ONNX_VERSION, FOUND_ONNX

from mct_quantizers.logger import Logger
from mct_quantizers.common.metadata import verify_and_init_metadata

if FOUND_TORCH:
    import torch

    def add_metadata(model: torch.nn.Module, metadata: Dict):
        """
        Init the metadata dictionary and verify its compliance, then add it to the model under
        the 'metadata' attribute.

        Args:
            model (Module): Pytorch model to add the metadata to.
            metadata (Dict): The metadata dictionary.

        Returns:
            The model with attribute 'metadata' that holds the metadata dictionary.

        Example:
            Adding author name and model version to model's metadata

            >>> model_with_metadata = add_metadata(model, {'author': 'John Doe', 'model version': 3})
        """

        metadata = verify_and_init_metadata(metadata)
        if FRAMEWORK_VERSION not in metadata:
            metadata[FRAMEWORK_VERSION] = torch.__version__
        model.metadata = metadata
        return model

    def get_metadata(model: torch.nn.Module) -> Dict:
        """
        Get the metadata dictionary from model.

        Args:
            model (Module): Pytorch model to extract metadata from.

        Returns:
            The model's the metadata dictionary.

        Example:
                Get model's metadata.

                >>> metadata = get_metadata(model)
        """
        return getattr(model, 'metadata', {})

else:
    def add_metadata(model,
                     metadata):
            Logger.critical('Installing pytorch is mandatory '
                            'when using add_metadata. '
                            'Could not find Pytorch package.')  # pragma: no cover

    def get_metadata(model):
            Logger.critical('Installing pytorch is mandatory '
                            'when using get_metadata. '
                            'Could not find Pytorch package.')  # pragma: no cover


if FOUND_ONNX:
    import onnx

    def add_onnx_metadata(model: onnx.ModelProto, metadata: Dict):
        """
        Init the metadata dictionary and verify its compliance, then add it to the model metadata_props.
        Metadata values have to be of byte type.

        Args:
            model (ModelProto): onnx model to add the metadata to.
            metadata (Dict): The metadata dictionary.

        Returns:
            The model with metadata dictionary.

        Example:
            Adding author name and model version to an onnx model's metadata

            >>> onnx_model_with_metadata = add_onnx_metadata(onnx_model, {'author': 'Jane Doe', 'model version': 3.5})
        """
        metadata = verify_and_init_metadata(metadata)
        if ONNX_VERSION not in metadata:
            metadata[ONNX_VERSION] = onnx.__version__

        for k, v in metadata.items():
            meta = model.metadata_props.add()
            if not isinstance(v, (bytes, str)):
                Logger.critical(f"ONNX metadata must be of byte type, but {v} has type {type(v)}")
            meta.key, meta.value = k, v
        return model


    def get_onnx_metadata(model: onnx.ModelProto) -> Dict:
        """
        Get the metadata dictionary from model.

        Args:
            model (ModelProto): onnx model to add the metadata to.

        Returns:
            The model's the metadata dictionary.

        Example:
                Get model's metadata.

                >>> metadata = get_onnx_metadata(model)
        """
        return {prop.key: prop.value for prop in model.metadata_props}

else:
    def add_onnx_metadata(model,
                          metadata):
            Logger.critical('Installing onnx is mandatory '
                            'when using add_onnx_metadata. '
                            'Could not find onnx package.')  # pragma: no cover

    def get_onnx_metadata(model):
            Logger.critical('Installing onnx is mandatory '
                            'when using get_onnx_metadata. '
                            'Could not find onnx package.')  # pragma: no cover
